#ifndef BUNDLE_H_INCLUDED
#define BUNDLE_H_INCLUDED

#include <stdint.h>

#include <upcn.h>

#define BUNDLE_INVALID_ID 0
typedef uint16_t bundleid_t;

/**
 * References a Bundle Protocol endpoint ID by two offsets into the dictionary
 * of the containing bundle
 */
struct eid_reference {
	uint16_t scheme_offset;
	uint16_t ssp_offset;
};

enum bundle_proc_flags {
	/* General */
	BUNDLE_FLAG_NONE = 0x00,
	BUNDLE_FLAG_IS_FRAGMENT = 0x01,
	BUNDLE_FLAG_ADMINISTRATIVE_RECORD = 0x02,
	BUNDLE_FLAG_MUST_NOT_FRAGMENT = 0x04,
	BUNDLE_FLAG_CUSTODY_TRANSFER_REQUESTED = 0x08,
	BUNDLE_FLAG_SINGLETON_ENDPOINT = 0x10,
	BUNDLE_FLAG_ACKNOWLEDGEMENT_REQUESTED = 0x20,
	/* Priority */
	BUNDLE_FLAG_NORMAL_PRIORITY = 0x080,
	BUNDLE_FLAG_EXPEDITED_PRIORITY = 0x100,
	/* Reports */
	BUNDLE_FLAG_REPORT_RECEPTION = 0x4000,
	BUNDLE_FLAG_REPORT_CUSTODY_ACCEPTANCE = 0x8000,
	BUNDLE_FLAG_REPORT_FORWARDING = 0x10000,
	BUNDLE_FLAG_REPORT_DELIVERY = 0x20000,
	BUNDLE_FLAG_REPORT_DELETION = 0x40000
};

enum bundle_retention_constraints {
	BUNDLE_RET_CONSTRAINT_NONE = 0x00,
	BUNDLE_RET_CONSTRAINT_DISPATCH_PENDING = 0x01,
	BUNDLE_RET_CONSTRAINT_FORWARD_PENDING = 0x02,
	BUNDLE_RET_CONSTRAINT_CUSTODY_ACCEPTED = 0x04,
	BUNDLE_RET_CONSTRAINT_REASSEMBLY_PENDING = 0x08,
	BUNDLE_RET_CONSTRAINT_FLAG_OWN = 0x80
};

enum bundle_block_type {
	BUNDLE_BLOCK_TYPE_PAYLOAD = 0x01
};

enum bundle_block_flags {
	BUNDLE_BLOCK_FLAG_NONE = 0x00,
	BUNDLE_BLOCK_FLAG_MUST_BE_REPLICATED = 0x01,
	BUNDLE_BLOCK_FLAG_REPORT_IF_UNPROC = 0x02,
	BUNDLE_BLOCK_FLAG_DELETE_BUNDLE_IF_UNPROC = 0x04,
	BUNDLE_BLOCK_FLAG_LAST_BLOCK = 0x08,
	BUNDLE_BLOCK_FLAG_DISCARD_IF_UNPROC = 0x10,
	BUNDLE_BLOCK_FLAG_FWD_UNPROC = 0x20,
	BUNDLE_BLOCK_FLAG_HAS_EID_REF_FIELD = 0x40
};

struct bundle_block {
	enum bundle_block_type type;
	enum bundle_block_flags flags;
	uint16_t eid_ref_count;
	struct eid_reference *eid_refs;
	uint32_t length;
	uint8_t *data;
};

struct bundle_block_list {
	struct bundle_block *data;
	struct bundle_block_list *next;
};

struct bundle {
	bundleid_t id;

	uint8_t protocol_version;

	enum bundle_proc_flags proc_flags;
	enum bundle_retention_constraints ret_constraints;

	struct eid_reference destination_eid;
	struct eid_reference source_eid;
	struct eid_reference report_eid;
	struct eid_reference custodian_eid;

	uint64_t creation_timestamp;
	uint16_t sequence_number;
	uint64_t lifetime;

	uint16_t dict_length;
	char *dict;

	uint32_t fragment_offset;
	uint32_t total_adu_length;

	uint16_t primary_prefix_length;
	uint16_t primary_block_length;

	struct bundle_block_list *blocks;
	struct bundle_block *payload_block;
};

enum bundle_administrative_record_type {
	BUNDLE_AR_STATUS_REPORT  = 0x1,
	BUNDLE_AR_CUSTODY_SIGNAL = 0x2
};

enum bundle_administrative_record_flags {
	BUNDLE_AR_FLAG_FRAGMENT = 0x1
};

enum bundle_status_report_status_flags {
	BUNDLE_SR_FLAG_BUNDLE_RECEIVED  = 0x01,
	BUNDLE_SR_FLAG_CUSTODY_ACCEPTED = 0x02,
	BUNDLE_SR_FLAG_BUNDLE_FORWARDED = 0x04,
	BUNDLE_SR_FLAG_BUNDLE_DELIVERED = 0x08,
	BUNDLE_SR_FLAG_BUNDLE_DELETED   = 0x10
};

enum bundle_status_report_reason {
	BUNDLE_SR_REASON_NO_INFO = 0x00,
	BUNDLE_SR_REASON_LIFETIME_EXPIRED = 0x01,
	BUNDLE_SR_REASON_FORWARDED_UNIDIRECTIONAL = 0x02,
	BUNDLE_SR_REASON_TRANSMISSION_CANCELED = 0x03,
	BUNDLE_SR_REASON_DEPLETED_STORAGE = 0x04,
	BUNDLE_SR_REASON_DEST_EID_UNINTELLIGIBLE = 0x05,
	BUNDLE_SR_REASON_NO_KNOWN_ROUTE = 0x06,
	BUNDLE_SR_REASON_NO_TIMELY_CONTACT = 0x07,
	BUNDLE_SR_REASON_BLOCK_UNINTELLIGIBLE = 0x08
};

struct bundle_status_report {
	enum bundle_status_report_status_flags status;
	enum bundle_status_report_reason reason;
};

enum bundle_custody_signal_reason {
	BUNDLE_CS_REASON_NO_INFO = 0x00,
	BUNDLE_CS_REASON_REDUNDANT_RECEPTION = 0x03,
	BUNDLE_CS_REASON_DEPLETED_STORAGE = 0x04,
	BUNDLE_CS_REASON_DEST_EID_UNINTELLIGIBLE = 0x05,
	BUNDLE_CS_REASON_NO_KNOWN_ROUTE = 0x06,
	BUNDLE_CS_REASON_NO_TIMELY_CONTACT = 0x07,
	BUNDLE_CS_REASON_BLOCK_UNINTELLIGIBLE = 0x08
};

struct bundle_custody_signal {
	int custody_transfer_succeeded;
	enum bundle_custody_signal_reason reason;
};

struct bundle_administrative_record {
	enum bundle_administrative_record_type type;
	enum bundle_administrative_record_flags flags;

	struct bundle_status_report *status_report;
	struct bundle_custody_signal *custody_signal;

	uint32_t fragment_offset;
	uint32_t fragment_length;
	uint64_t event_timestamp;
	uint32_t event_nanoseconds;

	uint64_t bundle_creation_timestamp;
	uint16_t bundle_sequence_number;

	uint16_t bundle_source_eid_length;
	char *bundle_source_eid;
};

struct bundle_list {
	struct bundle *data;
	struct bundle_list *next;
};

enum bundle_routing_priority {
	BUNDLE_RPRIO_LOW    = 0,
	BUNDLE_RPRIO_NORMAL = 1,
	BUNDLE_RPRIO_HIGH   = 2,
	BUNDLE_RPRIO_MAX
};

#define bundle_get_routing_priority(bundle) \
	((HAS_FLAG(bundle->ret_constraints, BUNDLE_RET_CONSTRAINT_FLAG_OWN) \
		|| HAS_FLAG(bundle->ret_constraints, \
			BUNDLE_RET_CONSTRAINT_CUSTODY_ACCEPTED) \
		|| HAS_FLAG(bundle->proc_flags, \
			BUNDLE_FLAG_EXPEDITED_PRIORITY)) \
				? BUNDLE_RPRIO_HIGH \
	: (HAS_FLAG(bundle->proc_flags, BUNDLE_FLAG_NORMAL_PRIORITY) \
		? BUNDLE_RPRIO_NORMAL \
	: BUNDLE_RPRIO_LOW))

#define bundle_is_fragmented(bundle) \
	((bundle->proc_flags & BUNDLE_FLAG_IS_FRAGMENT) != 0)

#define bundle_get_expiration_time(bundle) \
	(bundle->creation_timestamp + bundle->lifetime)

struct bundle *bundle_init();
void bundle_free_dynamic_parts(struct bundle *bundle);
void bundle_reset(struct bundle *bundle);
void bundle_free(struct bundle *bundle);
void bundle_drop(struct bundle *bundle);
char *bundle_read_eid(struct bundle *bundle, struct eid_reference eidref);
void bundle_copy_headers(struct bundle *to, struct bundle *from);
void bundle_recalculate_header_length(struct bundle *bundle);

uint32_t bundle_sizeof(struct bundle *bundle);
/* TODO: Maybe cache this in bundle struct... */
uint32_t bundle_block_get_size(struct bundle_block *block);
uint32_t bundle_get_serialized_size(struct bundle *bundle);
uint32_t bundle_get_serialized_size_without_payload(struct bundle *bundle);
uint32_t bundle_get_first_fragment_min_size(struct bundle *bundle);
uint32_t bundle_get_mid_fragment_min_size(struct bundle *bundle);
uint32_t bundle_get_last_fragment_min_size(struct bundle *bundle);

int bundle_eid_equals(
	const struct bundle *bundle, const struct eid_reference ref,
	const char *scheme, const char *ssp);
int bundle_eid_equals_string(
	const struct bundle *bundle, const struct eid_reference ref,
	const char *eid, size_t eid_len);

struct bundle_block *bundle_block_create(enum bundle_block_type t);
struct bundle_block_list *bundle_block_entry_create(struct bundle_block *b);
void bundle_block_free(struct bundle_block *b);
struct bundle_block_list *bundle_block_entry_free(struct bundle_block_list *e);
struct bundle_block *bundle_block_dup(struct bundle_block *b);
struct bundle_block_list *bundle_block_entry_dup(struct bundle_block_list *e);
struct bundle_block_list *bundle_block_list_dup(struct bundle_block_list *e);

#ifdef INCLUDE_ADVANCED_COMM

void bundle_print(struct bundle *bundle);
void bundle_print_stack_conserving(struct bundle *bundle);

#endif /* INCLUDE_ADVANCED_COMM */

#endif /* BUNDLE_H_INCLUDED */
