#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#include <mini-printf.h>

#include <upcn.h>
#include <bundle.h>
#include <sdnv.h>
#include <eidManager.h>

static inline void bundle_reset_internal(struct bundle *bundle)
{
	if (bundle == NULL)
		return;
	bundle->id = BUNDLE_INVALID_ID;
	bundle->protocol_version = 0x06;
	bundle->proc_flags = BUNDLE_FLAG_NONE;
	bundle->ret_constraints = BUNDLE_RET_CONSTRAINT_NONE;
	bundle->destination_eid.scheme_offset = 0;
	bundle->destination_eid.ssp_offset = 0;
	bundle->source_eid.scheme_offset = 0;
	bundle->source_eid.ssp_offset = 0;
	bundle->report_eid.scheme_offset = 0;
	bundle->report_eid.ssp_offset = 0;
	bundle->custodian_eid.scheme_offset = 0;
	bundle->custodian_eid.ssp_offset = 0;
	bundle->creation_timestamp = 0;
	bundle->sequence_number = 0;
	bundle->lifetime = 0;
	bundle->dict = NULL;
	bundle->dict_length = 0;
	bundle->fragment_offset = 0;
	bundle->total_adu_length = 0;
	bundle->primary_prefix_length = 0;
	bundle->primary_block_length = 0;
	bundle->blocks = NULL;
	bundle->payload_block = NULL;
}

struct bundle *bundle_init()
{
	struct bundle *bundle;

	bundle = malloc(sizeof(struct bundle));
	bundle_reset_internal(bundle);
	return bundle;
}

inline void bundle_free_dynamic_parts(struct bundle *bundle)
{
	ASSERT(bundle != NULL);
	if (bundle->dict != NULL) {
		free(bundle->dict);
		bundle->dict = NULL;
	}
	while (bundle->blocks != NULL)
		bundle->blocks = bundle_block_entry_free(bundle->blocks);
}

void bundle_reset(struct bundle *bundle)
{
	bundle_free_dynamic_parts(bundle);
	bundle_reset_internal(bundle);
}

void bundle_free(struct bundle *bundle)
{
	if (bundle == NULL)
		return;
	bundle_free_dynamic_parts(bundle);
	free(bundle);
}

void bundle_drop(struct bundle *bundle)
{
	ASSERT(bundle->ret_constraints == BUNDLE_RET_CONSTRAINT_NONE);
	bundle_free(bundle);
}

char *bundle_read_eid(struct bundle *bundle, struct eid_reference eidref)
{
	/* Both parts can have up to 1024 bytes including the trailing zero */
	/* as well as the colon between the parts... */
	int max_buf = 2049;
	char *eid_buffer;
	int str_len;

	/* ...but they can't be longer than the dictionary. */
	if ((bundle->dict_length + 1) < max_buf)
		max_buf = bundle->dict_length + 1;
	eid_buffer = malloc(sizeof(char) * max_buf);
	if (eid_buffer == NULL)
		return NULL;
	str_len = mini_snprintf(eid_buffer, max_buf, "%s:%s",
		(char *)(bundle->dict + eidref.scheme_offset),
		(char *)(bundle->dict + eidref.ssp_offset)) + 1;
	if (str_len <= 0) {
		free(eid_buffer);
		return NULL;
	} else if (str_len < max_buf) {
		eid_buffer = realloc(eid_buffer, sizeof(char) * str_len);
	}
	return eidmanager_alloc_ref(eid_buffer, 1);
}

void bundle_copy_headers(struct bundle *to, struct bundle *from)
{
	memcpy(to, from, sizeof(struct bundle));
	to->dict = malloc(from->dict_length + 1);
	if (to->dict != NULL)
		memcpy(to->dict, from->dict, from->dict_length + 1);
	to->blocks = NULL;
	to->payload_block = NULL;
}

void bundle_recalculate_header_length(struct bundle *bundle)
{
	bundle->primary_block_length
		= sdnv_get_size_u32(bundle->destination_eid.scheme_offset)
		+ sdnv_get_size_u32(bundle->destination_eid.ssp_offset)
		+ sdnv_get_size_u32(bundle->source_eid.scheme_offset)
		+ sdnv_get_size_u32(bundle->source_eid.ssp_offset)
		+ sdnv_get_size_u32(bundle->report_eid.scheme_offset)
		+ sdnv_get_size_u32(bundle->report_eid.ssp_offset)
		+ sdnv_get_size_u32(bundle->custodian_eid.scheme_offset)
		+ sdnv_get_size_u32(bundle->custodian_eid.ssp_offset)
		+ sdnv_get_size_u32(bundle->creation_timestamp)
		+ sdnv_get_size_u32(bundle->sequence_number)
		+ sdnv_get_size_u32(bundle->lifetime)
		+ sdnv_get_size_u32(bundle->dict_length)
		+ bundle->dict_length
		+ (HAS_FLAG(bundle->proc_flags, BUNDLE_FLAG_IS_FRAGMENT)
			? (sdnv_get_size_u32(bundle->fragment_offset)
				+ sdnv_get_size_u32(bundle->total_adu_length))
			: 0);
	bundle->primary_prefix_length = 1
		+ sdnv_get_size_u32(bundle->proc_flags)
		+ sdnv_get_size_u32(bundle->primary_block_length);
}

uint32_t bundle_sizeof(struct bundle *bundle)
{
	uint32_t bs = 0;
	struct bundle_block_list *cur = bundle->blocks;

	while (cur != NULL) {
		bs += sizeof(struct bundle_block_list)
			+ sizeof(struct bundle_block)
			+ cur->data->length;
		cur = cur->next;
	}
	return bs + sizeof(struct bundle) + bundle->dict_length;
}

uint32_t bundle_block_get_size(struct bundle_block *block)
{
	int i;
	uint16_t eid_size = 0;

	if (HAS_FLAG(block->flags, BUNDLE_BLOCK_FLAG_HAS_EID_REF_FIELD)) {
		eid_size += sdnv_get_size_u16(block->eid_ref_count);
		for (i = 0; i < block->eid_ref_count; i++) {
			eid_size += sdnv_get_size_u16(
				block->eid_refs[i].scheme_offset);
			eid_size += sdnv_get_size_u16(
				block->eid_refs[i].ssp_offset);
		}
	}
	return 1
		+ sdnv_get_size_u32(block->flags)
		+ sdnv_get_size_u32(block->length)
		+ block->length
		+ eid_size;
}

uint32_t bundle_get_serialized_size(struct bundle *bundle)
{
	uint32_t s = 0;
	struct bundle_block_list *cur = bundle->blocks;

	while (cur != NULL) {
		s += bundle_block_get_size(cur->data);
		cur = cur->next;
	}
	return bundle->primary_prefix_length + bundle->primary_block_length + s;
}

uint32_t bundle_get_serialized_size_without_payload(struct bundle *bundle)
{
	uint32_t s = 0;
	struct bundle_block_list *cur = bundle->blocks;

	while (cur != NULL) {
		if (cur->data->type != BUNDLE_BLOCK_TYPE_PAYLOAD)
			s += bundle_block_get_size(cur->data);
		cur = cur->next;
	}
	return bundle->primary_prefix_length + bundle->primary_block_length + s;
}

uint32_t bundle_get_first_fragment_min_size(struct bundle *bundle)
{
	uint32_t s = 0;
	int_fast8_t payload_reached = 0;
	struct bundle_block_list *cur = bundle->blocks;

	while (cur != NULL) {
		if (cur->data->type == BUNDLE_BLOCK_TYPE_PAYLOAD) {
			payload_reached = 1;
			/* Payload header size */
			s += 1  + sdnv_get_size_u32(cur->data->flags)
				+ sdnv_get_size_u32(cur->data->length);
		} else if (!payload_reached
			|| HAS_FLAG(cur->data->flags,
				BUNDLE_BLOCK_FLAG_MUST_BE_REPLICATED)
		) {
			s += bundle_block_get_size(cur->data);
		}
		cur = cur->next;
	}
	/* Add sizes of total ADU len (= PL) + 1 byte for offset 0 */
	if (!HAS_FLAG(bundle->proc_flags, BUNDLE_FLAG_IS_FRAGMENT))
		s += sdnv_get_size_u32(bundle->payload_block->length) + 1;
	return bundle->primary_prefix_length + bundle->primary_block_length + s;
}

uint32_t bundle_get_mid_fragment_min_size(struct bundle *bundle)
{
	uint32_t s = 0;
	struct bundle_block_list *cur = bundle->blocks;

	while (cur != NULL) {
		if (cur->data->type == BUNDLE_BLOCK_TYPE_PAYLOAD) {
			/* Payload header size */
			s += 1  + sdnv_get_size_u32(cur->data->flags)
				+ sdnv_get_size_u32(cur->data->length);
		} else if (HAS_FLAG(cur->data->flags,
			BUNDLE_BLOCK_FLAG_MUST_BE_REPLICATED)
		) {
			s += bundle_block_get_size(cur->data);
		}
		cur = cur->next;
	}
	/* Add sizes of total ADU len/offset, assume offset SDNV has PL len */
	if (!HAS_FLAG(bundle->proc_flags, BUNDLE_FLAG_IS_FRAGMENT))
		s += sdnv_get_size_u32(bundle->payload_block->length) * 2;
	return bundle->primary_prefix_length + bundle->primary_block_length + s;
}

uint32_t bundle_get_last_fragment_min_size(struct bundle *bundle)
{
	uint32_t s = 0;
	int_fast8_t payload_reached = 0;
	struct bundle_block_list *cur = bundle->blocks;

	while (cur != NULL) {
		if (cur->data->type == BUNDLE_BLOCK_TYPE_PAYLOAD) {
			payload_reached = 1;
			/* Payload header size */
			s += 1  + sdnv_get_size_u32(cur->data->flags)
				+ sdnv_get_size_u32(cur->data->length);
		} else if (
			payload_reached
			|| HAS_FLAG(cur->data->flags,
				BUNDLE_BLOCK_FLAG_MUST_BE_REPLICATED)
		) {
			s += bundle_block_get_size(cur->data);
		}
		cur = cur->next;
	}
	/* Add sizes of total ADU len/offset, assume offset SDNV has PL len */
	if (!HAS_FLAG(bundle->proc_flags, BUNDLE_FLAG_IS_FRAGMENT))
		s += sdnv_get_size_u32(bundle->payload_block->length) * 2;
	return bundle->primary_prefix_length + bundle->primary_block_length + s;
}

int bundle_eid_equals(
	const struct bundle *bundle, const struct eid_reference ref,
	const char *scheme, const char *ssp)
{
	return (strcmp(bundle->dict + ref.scheme_offset, scheme) == 0 &&
		strcmp(bundle->dict + ref.ssp_offset, ssp) == 0);
}

int bundle_eid_equals_string(
	const struct bundle *bundle, const struct eid_reference ref,
	const char *eid, size_t eid_len)
{
	size_t scheme_len = strlen(bundle->dict + ref.scheme_offset);

	return (strncmp(bundle->dict + ref.scheme_offset, eid, scheme_len) == 0
		&& strncmp(bundle->dict + ref.ssp_offset,
			eid + scheme_len + 1, eid_len - scheme_len - 1) == 0);
}

struct bundle_block *bundle_block_create(enum bundle_block_type t)
{
	struct bundle_block *block = malloc(sizeof(struct bundle_block));

	if (block == NULL)
		return NULL;
	block->type = t;
	block->flags = BUNDLE_BLOCK_FLAG_NONE;
	block->eid_ref_count = 0;
	block->eid_refs = NULL;
	block->length = 0;
	block->data = NULL;
	return block;
}

struct bundle_block_list *bundle_block_entry_create(struct bundle_block *b)
{
	struct bundle_block_list *entry;

	if (b == NULL)
		return NULL;
	entry = malloc(sizeof(struct bundle_block_list));
	if (entry == NULL)
		return NULL;
	entry->data = b;
	entry->next = NULL;
	return entry;
}

void bundle_block_free(struct bundle_block *b)
{
	if (b != NULL) {
		if (b->eid_refs != NULL)
			free(b->eid_refs);
		if (b->data != NULL)
			free(b->data);
		free(b);
	}
}

struct bundle_block_list *bundle_block_entry_free(struct bundle_block_list *e)
{
	struct bundle_block_list *next;

	ASSERT(e != NULL);
	next = e->next;
	bundle_block_free(e->data);
	free(e);
	return next;
}

struct bundle_block *bundle_block_dup(struct bundle_block *b)
{
	struct bundle_block *dup;
	size_t s;

	ASSERT(b != NULL);
	dup = malloc(sizeof(struct bundle_block));
	if (dup == NULL)
		return NULL;
	memcpy(dup, b, sizeof(struct bundle_block));
	if (dup->eid_refs != NULL) {
		s = sizeof(struct eid_reference) * b->eid_ref_count;
		dup->eid_refs = malloc(s);
		if (dup->eid_refs == NULL)
			goto err;
		memcpy(dup->eid_refs, b->eid_refs, s);
	}
	dup->data = malloc(b->length);
	if (dup->data == NULL)
		goto err;
	memcpy(dup->data, b->data, b->length);
	return dup;
err:
	if (dup->eid_refs != NULL)
		free(dup->eid_refs);
	free(dup);
	return NULL;
}

struct bundle_block_list *bundle_block_entry_dup(struct bundle_block_list *e)
{
	struct bundle_block *dup;
	struct bundle_block_list *result;

	ASSERT(e != NULL);
	dup = bundle_block_dup(e->data);
	if (dup == NULL)
		return NULL;
	result = bundle_block_entry_create(dup);
	if (result == NULL)
		bundle_block_free(dup);
	return result;
}

struct bundle_block_list *bundle_block_list_dup(struct bundle_block_list *e)
{
	struct bundle_block_list *dup = NULL, **cur = &dup;

	while (e != NULL) {
		*cur = bundle_block_entry_dup(e);
		if (!*cur) {
			while (dup)
				dup = bundle_block_entry_free(dup);
			return NULL;
		}
		cur = &(*cur)->next;
		e = e->next;
	}
	return dup;
}

#ifdef INCLUDE_ADVANCED_COMM

/* This possibly needs more than 600 bytes of stack space */
void bundle_print(struct bundle *bundle)
{
	char *tmp;

	comm_lock();
	comm_printf("\nBUNDLE\n======\nPointer:   %p\n", bundle);
	comm_printf("Bdl. ID:   %d\n", bundle->id);
	comm_printf("Headr.Sz.: %d + %d\n",
		bundle->primary_block_length, bundle->primary_prefix_length);
	comm_printf("Seria.Sz.: %d\n", bundle_get_serialized_size(bundle));
	comm_printf("Frag1.Sz.: %d\n",
		bundle_get_first_fragment_min_size(bundle));
	comm_printf("Flags:     %#x\n", bundle->proc_flags);
	comm_printf("Dest.:     %s\n",
		tmp = bundle_read_eid(bundle, bundle->destination_eid));
	free(tmp);
	comm_printf("Source:    %s\n",
		tmp = bundle_read_eid(bundle, bundle->source_eid));
	free(tmp);
	comm_printf("Report to: %s\n",
		tmp = bundle_read_eid(bundle, bundle->report_eid));
	free(tmp);
	comm_printf("Custodian: %s\n",
		tmp = bundle_read_eid(bundle, bundle->custodian_eid));
	free(tmp);
	comm_printf("Timestamp: %"PRIu32"\n", bundle->creation_timestamp);
	comm_printf("SequenceN: %"PRIu32"\n", bundle->sequence_number);
	comm_printf("Lifetime:  %"PRIu32"\n", bundle->lifetime);
	comm_printf("Dict.Len.: %"PRIu32"\n", bundle->dict_length);
	comm_printf("Frag.Off.: %"PRIu32"\n", bundle->fragment_offset);
	comm_printf("Total ADU: %"PRIu32"\n", bundle->total_adu_length);
	comm_printf("Pay.Flags: %#x\n", bundle->payload_block->flags);
	comm_printf("Pay.Len.:  %"PRIu32"\n", bundle->payload_block->length);
	comm_unlock();
}

/* This will use ~80 bytes of stack space */
void bundle_print_stack_conserving(struct bundle *bundle)
{
	char buffer[16];

	comm_lock();
	comm_write("\nBUNDLE\n======\nPointer:   ");
	sprintu32x(buffer, (uint32_t)bundle);
	comm_write(buffer);
	comm_write("\nBdl. ID:   ");
	sprintu32(buffer, (uint32_t)bundle->id);
	comm_write(buffer);
	comm_write("\nHeadr.Sz.: ");
	sprintu32(buffer, bundle->primary_block_length);
	comm_write(buffer);
	comm_write(" + ");
	sprintu32(buffer, bundle->primary_prefix_length);
	comm_write(buffer);
	comm_write("\nSeria.Sz.: ");
	sprintu32(buffer, bundle_get_serialized_size(bundle));
	comm_write(buffer);
	comm_write("\nFlags:     ");
	sprintu32(buffer, bundle->proc_flags);
	comm_write(buffer);
	comm_write("\nTimestamp: ");
	sprintu64(buffer, bundle->creation_timestamp);
	comm_write(buffer);
	comm_write("\nFrag.Off.: ");
	sprintu32(buffer, bundle->fragment_offset);
	comm_write(buffer);
	comm_write("\nTotal ADU: ");
	sprintu32(buffer, bundle->total_adu_length);
	comm_write(buffer);
	comm_write("\nPay.Flags: ");
	sprintu32x(buffer, bundle->payload_block->flags);
	comm_write(buffer);
	comm_write("\nPay.Len.:  ");
	sprintu32(buffer, bundle->payload_block->length);
	comm_write(buffer);
	comm_write("\n");
	comm_unlock();
}

#endif /* INCLUDE_ADVANCED_COMM */
