#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <upcn.h>
#include <bundle.h>
#include <groundStation.h>
#include <bundleFragmenter.h>

static struct bundle *create_new_fragment(
	struct bundle *prototype, uint8_t init_pl);
static enum upcn_result replicate_blocks(
	struct bundle *first, struct bundle *second);

struct bundle *bundlefragmenter_initialize_first_fragment(struct bundle *input)
{
	struct bundle *result;
	struct bundle_block_list *cur_block;

	ASSERT(input != NULL);
	result = create_new_fragment(input, 0);
	if (result == NULL)
		return NULL;
	cur_block = bundle_block_list_dup(input->blocks);
	result->blocks = cur_block;
	while (cur_block != NULL) {
		if (cur_block->data->type == BUNDLE_BLOCK_TYPE_PAYLOAD) {
			result->payload_block = cur_block->data;
			break;
		}
		cur_block = cur_block->next;
	}
	if (result->payload_block == NULL) {
		bundle_free(result);
		return NULL;
	}
	bundle_recalculate_header_length(result);
	return result;
}

struct bundle *bundlefragmenter_fragment_bundle(
	struct bundle *working_bundle, uint64_t first_max)
{
	struct bundle *remainder;
	struct bundle_block_list *cur_block;
	uint32_t first_payload_length;

	/* Determine length of first fragment's payload */
	first_payload_length = first_max
		- bundle_get_first_fragment_min_size(working_bundle);
	/* Don't fragment? */
	if (first_payload_length >= working_bundle->payload_block->length)
		return working_bundle;
	/* Construct second fragment */
	remainder = create_new_fragment(working_bundle, 1);
	if (remainder == NULL)
		return NULL;
	/* Set total ADU length if not already fragmented */
	if (!HAS_FLAG(working_bundle->proc_flags, BUNDLE_FLAG_IS_FRAGMENT))
		working_bundle->total_adu_length
			= working_bundle->payload_block->length;
	/* Set the payload block's properties */
	remainder->payload_block->length
		= working_bundle->payload_block->length - first_payload_length;
	remainder->payload_block->data
		= malloc(remainder->payload_block->length);
	if (remainder->payload_block->data == NULL) {
		bundle_free(remainder);
		return NULL;
	}
	memcpy(
		remainder->payload_block->data,
		working_bundle->payload_block->data + first_payload_length,
		remainder->payload_block->length
	);
	/* Find PL block position in working bundle */
	/* Add following blocks to remainder */
	cur_block = working_bundle->blocks;
	while (cur_block != NULL) {
		if (cur_block->data == working_bundle->payload_block) {
			/* PL block of first fragment is now last block */
			cur_block->data->flags |= BUNDLE_BLOCK_FLAG_LAST_BLOCK;
			remainder->blocks->next = cur_block->next;
			cur_block->next = NULL;
			break;
		}
		cur_block = cur_block->next;
	}
	/* Shorten first fragment's PL block */
	working_bundle->payload_block->data
		= realloc(working_bundle->payload_block->data,
			first_payload_length);
	ASSERT(working_bundle->payload_block->data != NULL);
	/* Set correct lengths and offsets */
	working_bundle->payload_block->length = first_payload_length;
	remainder->fragment_offset
		= working_bundle->fragment_offset + first_payload_length;
	/* Handle blocks that must be replicated in every fragment */
	replicate_blocks(working_bundle, remainder);
	/* Set fragment flag */
	working_bundle->proc_flags |= BUNDLE_FLAG_IS_FRAGMENT;
	/* Recalc sizes */
	bundle_recalculate_header_length(working_bundle);
	bundle_recalculate_header_length(remainder);
	/* Return created fragment */
	return remainder;
}

/* creates a new fragment for the given bundle */
static struct bundle *create_new_fragment(
	struct bundle *prototype, uint8_t init_pl)
{
	struct bundle *fragment;
	struct bundle_block *fragment_pl;

	fragment = malloc(sizeof(struct bundle));
	if (fragment == NULL)
		return NULL;
	bundle_copy_headers(fragment, prototype);
	/* Reset the ID */
	fragment->id = BUNDLE_INVALID_ID;
	/* Set total ADU length if not already fragmented */
	if (!HAS_FLAG(prototype->proc_flags, BUNDLE_FLAG_IS_FRAGMENT))
		fragment->total_adu_length = prototype->payload_block->length;
	else
		fragment->total_adu_length = prototype->total_adu_length;
	/* Set fragment flag */
	fragment->proc_flags |= BUNDLE_FLAG_IS_FRAGMENT;
	if (init_pl) {
		/* Create new PL block */
		fragment_pl = bundle_block_create(BUNDLE_BLOCK_TYPE_PAYLOAD);
		if (fragment_pl == NULL) {
			free(fragment);
			return NULL;
		}
		fragment_pl->flags = prototype->payload_block->flags;
		/* Insert the payload block */
		fragment->blocks = bundle_block_entry_create(fragment_pl);
		if (fragment->blocks == NULL) {
			bundle_block_free(fragment_pl);
			free(fragment);
			return NULL;
		}
		/* Set the bundle's payload reference */
		fragment->payload_block = fragment_pl;
	}
	return fragment;
}

/* first and second have to be split @ payload block */
/* payload block has to be last block of first bundle and first of second */
static enum upcn_result replicate_blocks(
	struct bundle *first, struct bundle *second)
{
	struct bundle_block_list *list = NULL;
	struct bundle_block_list **item;
	struct bundle_block_list *entry;

	/* 1) Handle replicated blocks _BEFORE_ payload */
	list = NULL;
	item = &list;
	entry = first->blocks;
	while (entry != NULL) {
		if (
			HAS_FLAG(entry->data->flags,
				BUNDLE_BLOCK_FLAG_MUST_BE_REPLICATED)
			&& entry->data->type != BUNDLE_BLOCK_TYPE_PAYLOAD
		) {
			*item = bundle_block_entry_dup(entry);
			if (*item == NULL) {
				while (list != NULL)
					list = bundle_block_entry_free(list);
				return UPCN_FAIL;
			}
			item = &(*item)->next;
		}
		entry = entry->next;
	}
	/* Set first entry as PL block of 2nd fragment before we change it */
	entry = second->blocks; /* PL block is first till now */
	/* Prepend block list of 2nd fragment with replicated blocks */
	/* _BEFORE_ payload of 1st fragment */
	if (list != NULL) {
		(*item)->next = second->blocks;
		second->blocks = list;
	}
	/* 2) Handle replicated blocks _AFTER_ payload */
	list = NULL;
	item = &list;
	while (entry != NULL) {
		if (
			HAS_FLAG(entry->data->flags,
				 BUNDLE_BLOCK_FLAG_MUST_BE_REPLICATED)
			&& entry->data->type != BUNDLE_BLOCK_TYPE_PAYLOAD
		) {
			*item = bundle_block_entry_dup(entry);
			if (*item == NULL) {
				while (list != NULL)
					list = bundle_block_entry_free(list);
				return UPCN_FAIL;
			}
			item = &(*item)->next;
		}
		entry = entry->next;
	}
	/* Append replicated blocks _AFTER_ payload to list of first fragment */
	if (list != NULL) {
		/* Go to 1st fragment's last block (payload block) */
		entry = first->blocks;
		ASSERT(entry != NULL);
		while (entry->next != NULL)
			entry = entry->next;
		/* Remove "last_block" flag */
		entry->data->flags &= ~BUNDLE_BLOCK_FLAG_LAST_BLOCK;
		/* Append */
		entry->next = list;
	}
	return UPCN_OK;
}
