#include <stdlib.h>

#include <upcn.h>
#include <bundleParser.h>

#ifdef UPCN_LOCAL
#define bundle_storage_get_usage() (0)
#else /* UPCN_LOCAL */
#include <bundleStorageManager.h>
#endif /* UPCN_LOCAL */

struct parser *bundle_parser_init(
	struct bundle_parser *state, void (*send_callback)(struct bundle *))
{
	state->basedata = malloc(sizeof(struct parser));
	if (state->basedata == NULL)
		return NULL;
	state->send_callback = send_callback;
	state->bundle = NULL;
	state->basedata->status = PARSER_STATUS_ERROR;
	if (bundle_parser_reset(state) != UPCN_OK)
		return NULL;
	return state->basedata;
}

enum upcn_result bundle_parser_reset(struct bundle_parser *state)
{
	if (state->basedata->status == PARSER_STATUS_GOOD &&
			state->current_stage == PARSER_STAGE_VERSION)
		return UPCN_OK;

	state->basedata->status = PARSER_STATUS_GOOD;
	state->error = PARSER_ERROR_NONE;

	state->current_stage = PARSER_STAGE_VERSION;
	state->next_stage    = PARSER_STAGE_VERSION;
	state->current_block = PARSER_BLOCK_PRIMARY;

	sdnv_reset(&state->sdnv_state);

	state->primary_bytes_remaining = 0;
	state->cur_bytes_remaining = 0;
	state->current_index = 0;
	state->current_size = bundle_storage_get_usage()
		+ sizeof(struct bundle);
	state->last_block = 0;

	if (state->bundle != NULL)
		bundle_reset(state->bundle);
	else
		state->bundle = bundle_init();
	if (state->bundle == NULL)
		return UPCN_FAIL;

	state->current_block_entry = &state->bundle->blocks;

	return UPCN_OK;
}

static inline uint8_t bundle_parser_length_known(struct bundle_parser *state)
{
	return state->current_stage > PARSER_STAGE_BLOCK_LENGTH;
}

static inline int bundle_parser_data_done(struct bundle_parser *state)
{
	return state->cur_bytes_remaining == 0;
}

static inline int bundle_parser_is_last_block(struct bundle_parser *state)
{
	return state->last_block;
}

static inline void bundle_parser_begin_block(
	struct bundle_parser *state, uint8_t type)
{
	struct bundle_block *new = bundle_block_create(type);

	if (new == NULL) {
		state->basedata->status = PARSER_STATUS_ERROR;
		return;
	}
	if (type == BUNDLE_BLOCK_TYPE_PAYLOAD) {
		state->bundle->payload_block = new;
		state->current_block = PARSER_BLOCK_PAYLOAD;
	} else {
		state->current_block = PARSER_BLOCK_EXTENSION;
	}
	*state->current_block_entry = bundle_block_entry_create(new);
	if (*state->current_block_entry == NULL) {
		bundle_block_free(new);
		state->basedata->status = PARSER_STATUS_ERROR;
		return;
	}
}

static inline void bundle_parser_end_block(struct bundle_parser *state)
{
	state->current_block_entry = &(*state->current_block_entry)->next;
}

static inline void bundle_parser_send_bundle(struct bundle_parser *state)
{
	struct bundle *ptr = state->bundle;

	state->bundle = NULL;
	if (state->send_callback == NULL)
		bundle_free(ptr);
	else
		state->send_callback(ptr);
}

static inline void bundle_parser_next(struct bundle_parser *state)
{
	switch (state->next_stage) {
	case PARSER_STAGE_PROC_FLAGS:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_LENGTH:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_DESTINATION_EID_SCHEME:
		state->primary_bytes_remaining =
			state->bundle->primary_block_length;
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_DESTINATION_EID_SSP:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_SOURCE_EID_SCHEME:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_SOURCE_EID_SSP:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_REPORT_EID_SCHEME:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_REPORT_EID_SSP:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_CUSTODIAN_EID_SCHEME:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_CUSTODIAN_EID_SSP:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_TIMESTAMP:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_SEQUENCE_NUM:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_LIFETIME:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_DICT_LENGTH:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_DICTIONARY:
		state->current_index = 0;
		state->cur_bytes_remaining =
			state->bundle->dict_length;
		state->current_size += state->cur_bytes_remaining + 1;
		if (state->current_size > BUNDLE_QUOTA) {
			state->basedata->status = PARSER_STATUS_ERROR;
		} else {
			state->bundle->dict =
				malloc(state->cur_bytes_remaining + 1);
			if (state->bundle->dict == NULL)
				state->basedata->status
					= PARSER_STATUS_ERROR;
		}
		break;
	case PARSER_STAGE_FRAGMENT_OFFSET:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_ADU_LENGTH:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_TYPE:
		state->current_block = PARSER_BLOCK_GENERIC;
		break;
	case PARSER_STAGE_BLOCK_FLAGS:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_EID_REF_CNT:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_EID_REF_SCH:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_EID_REF_SSP:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_DATA_LENGTH:
		sdnv_reset(&state->sdnv_state);
		break;
	case PARSER_STAGE_BLOCK_DATA:
		state->current_index = 0;
		state->cur_bytes_remaining =
			(*state->current_block_entry)->data->length;
		state->current_size += state->cur_bytes_remaining;
		if (state->current_size > BUNDLE_QUOTA) {
			state->basedata->status = PARSER_STATUS_ERROR;
		} else {
			(*state->current_block_entry)->data->data
				= malloc(state->cur_bytes_remaining);
			if ((*state->current_block_entry)
					->data->data == NULL)
				state->basedata->status = PARSER_STATUS_ERROR;
		}
	default:
		break;
	}

	state->current_stage = state->next_stage;

	if (state->current_stage == PARSER_STAGE_DONE) {
		state->basedata->status = PARSER_STATUS_DONE;
		bundle_parser_send_bundle(state);
	}
}

/**
 * Checks if the SDNV parser is done or has encountered an error
 *
 * In the former case, sets the parser to continue with the next stage,
 * in the latter case, "forwards" the error to the parser.
 */
static inline int bundle_parser_wait_for_sdnv(
	struct bundle_parser *state, enum bundle_parser_stage next_stage)
{
	switch (state->sdnv_state.status) {
	case SDNV_IN_PROGRESS:
		/* intentional, do nothing unless
		 * a termination state is reached
		 */
		break;
	case SDNV_DONE:
		state->next_stage = next_stage;
		return 1;
	case SDNV_ERROR:
		state->basedata->status = PARSER_STATUS_ERROR;
		state->error  = PARSER_ERROR_SDNV_FAILURE;
		break;
	}
	return 0;
}

void bundle_parser_read_byte(struct bundle_parser *state, uint8_t byte)
{
	if (state->current_block == PARSER_BLOCK_PRIMARY) {
		if (bundle_parser_length_known(state)
			&& (state->primary_bytes_remaining == 0)
		) {
			state->basedata->status = PARSER_STATUS_ERROR;
			state->error = PARSER_ERROR_BLOCK_LENGTH_EXHAUSTED;
			return;
		}
		state->primary_bytes_remaining--;
	}

	switch (state->current_stage) {
	case PARSER_STAGE_VERSION:
		state->bundle->primary_prefix_length++;
		state->bundle->protocol_version = byte;
		state->next_stage = PARSER_STAGE_PROC_FLAGS;
		break;
	case PARSER_STAGE_PROC_FLAGS:
		state->bundle->primary_prefix_length++;
		sdnv_read_u32(&state->sdnv_state,
			(uint32_t *)&state->bundle->proc_flags, byte);
		bundle_parser_wait_for_sdnv(
				state, PARSER_STAGE_BLOCK_LENGTH);
		break;
	case PARSER_STAGE_BLOCK_LENGTH:
		state->bundle->primary_prefix_length++;
		sdnv_read_u16(&state->sdnv_state,
				&state->bundle->primary_block_length, byte);
			bundle_parser_wait_for_sdnv(
				state, PARSER_STAGE_DESTINATION_EID_SCHEME);
		break;
	case PARSER_STAGE_DESTINATION_EID_SCHEME:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->destination_eid.scheme_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_DESTINATION_EID_SSP);
		break;
	case PARSER_STAGE_DESTINATION_EID_SSP:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->destination_eid.ssp_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_SOURCE_EID_SCHEME);
		break;
	case PARSER_STAGE_SOURCE_EID_SCHEME:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->source_eid.scheme_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_SOURCE_EID_SSP);
		break;
	case PARSER_STAGE_SOURCE_EID_SSP:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->source_eid.ssp_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_REPORT_EID_SCHEME);
		break;
	case PARSER_STAGE_REPORT_EID_SCHEME:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->report_eid.scheme_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_REPORT_EID_SSP);
		break;
	case PARSER_STAGE_REPORT_EID_SSP:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->report_eid.ssp_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_CUSTODIAN_EID_SCHEME);
		break;
	case PARSER_STAGE_CUSTODIAN_EID_SCHEME:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->custodian_eid.scheme_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_CUSTODIAN_EID_SSP);
		break;
	case PARSER_STAGE_CUSTODIAN_EID_SSP:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->custodian_eid.ssp_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_TIMESTAMP);
		break;
	case PARSER_STAGE_TIMESTAMP:
		sdnv_read_u64(
			&state->sdnv_state,
			&state->bundle->creation_timestamp,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_SEQUENCE_NUM);
		break;
	case PARSER_STAGE_SEQUENCE_NUM:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->sequence_number,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_LIFETIME);
		break;
	case PARSER_STAGE_LIFETIME:
		sdnv_read_u64(
			&state->sdnv_state,
			&state->bundle->lifetime,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_DICT_LENGTH);
		break;
	case PARSER_STAGE_DICT_LENGTH:
		sdnv_read_u16(
			&state->sdnv_state,
			&state->bundle->dict_length,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_DICTIONARY);
		break;
	case PARSER_STAGE_DICTIONARY:
		--state->cur_bytes_remaining;
		((uint8_t *)state->bundle->dict)
			[state->current_index++] = byte;
		if (bundle_parser_data_done(state)) {
			((uint8_t *)state->bundle->dict)
				[state->current_index++] = '\0';
			if (bundle_is_fragmented(state->bundle))
				state->next_stage =
					PARSER_STAGE_FRAGMENT_OFFSET;
			else
				state->next_stage =
					PARSER_STAGE_BLOCK_TYPE;
		}
		break;
	case PARSER_STAGE_FRAGMENT_OFFSET:
		sdnv_read_u32(
			&state->sdnv_state,
			&state->bundle->fragment_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_ADU_LENGTH);
		break;
	case PARSER_STAGE_ADU_LENGTH:
		sdnv_read_u32(
			&state->sdnv_state,
			&state->bundle->total_adu_length,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_BLOCK_TYPE);
		break;
	case PARSER_STAGE_BLOCK_TYPE:
		state->next_stage = PARSER_STAGE_BLOCK_FLAGS;
		bundle_parser_begin_block(state, byte);
		break;
	case PARSER_STAGE_BLOCK_FLAGS:
		sdnv_read_u32(
			&state->sdnv_state,
			(uint32_t *)&(*state->current_block_entry)
				->data->flags, byte);
		if (bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_BLOCK_DATA_LENGTH)) {
			enum bundle_block_flags flags =
				(*state->current_block_entry)
					->data->flags;
			state->last_block = HAS_FLAG(flags,
				BUNDLE_BLOCK_FLAG_LAST_BLOCK);
			if (HAS_FLAG(flags,
				BUNDLE_BLOCK_FLAG_HAS_EID_REF_FIELD)
			) {
				state->next_stage =
					PARSER_STAGE_BLOCK_EID_REF_CNT;
			}
		}
		break;
	case PARSER_STAGE_BLOCK_EID_REF_CNT:
		sdnv_read_u16(
			&state->sdnv_state,
			&(*state->current_block_entry)
				->data->eid_ref_count, byte);
		if (bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_BLOCK_EID_REF_SCH)) {
			state->current_index = 0;
			size_t eidrefsz = sizeof(struct eid_reference)
				* (*state->current_block_entry)
					->data->eid_ref_count;
			state->current_size += eidrefsz;
			if (state->current_size > BUNDLE_QUOTA) {
				state->basedata->status
					= PARSER_STATUS_ERROR;
			} else {
				(*state->current_block_entry)->data
					->eid_refs = malloc(eidrefsz);
				if ((*state->current_block_entry)->data
					->eid_refs == NULL
				) {
					state->basedata->status
						= PARSER_STATUS_ERROR;
				}
			}
		}
		break;
	case PARSER_STAGE_BLOCK_EID_REF_SCH:
		sdnv_read_u16(
			&state->sdnv_state,
			&(*state->current_block_entry)
				->data->eid_refs[state->current_index]
				.scheme_offset,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_BLOCK_EID_REF_SSP);
		break;
	case PARSER_STAGE_BLOCK_EID_REF_SSP:
		sdnv_read_u16(
			&state->sdnv_state,
			&(*state->current_block_entry)
				->data->eid_refs[state->current_index]
				.ssp_offset,
			byte);
		if (bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_BLOCK_EID_REF_SSP)) {
			state->current_index++;
			if (state->current_index
				== (*state->current_block_entry)
					->data->eid_ref_count
			) {
				state->next_stage =
				PARSER_STAGE_BLOCK_DATA_LENGTH;
			}
		}
		break;
	case PARSER_STAGE_BLOCK_DATA_LENGTH:
		sdnv_read_u32(
			&state->sdnv_state,
			&(*state->current_block_entry)
				->data->length,
			byte);
		bundle_parser_wait_for_sdnv(
			state, PARSER_STAGE_BLOCK_DATA);
		break;
	case PARSER_STAGE_BLOCK_DATA:
		--state->cur_bytes_remaining;
		(*state->current_block_entry)->data->data
			[state->current_index++] = byte;
		if (bundle_parser_data_done(state)) {
			bundle_parser_end_block(state);
			if (bundle_parser_is_last_block(state))
				state->next_stage = PARSER_STAGE_DONE;
			else
				state->next_stage =
					PARSER_STAGE_BLOCK_TYPE;
		}
		break;
	default:
		state->basedata->status = PARSER_STATUS_ERROR;
		break;
	}

	if (state->current_stage != state->next_stage)
		bundle_parser_next(state);
}
