#include <stddef.h>
#include <stdint.h>

#include <config.h>
#include <upcnFlags.h>
#include <comm.h>

#ifdef INCLUDE_BOARD_LIB

#include <stm32f4xx.h>
#include <usbd_cdc_vcp.h>

#include <string.h>

#include <FreeRTOS.h>
#include <queue.h>
#include <semphr.h>

/* USB COMMUNICATION */

static SemaphoreHandle_t comm_semaphore;
static QueueHandle_t rx_queue, tx_queue;

static uint8_t tlvbuf[5];

#define COMM_BEGIN do { ; } \
	while (pdFALSE == xSemaphoreTake(comm_semaphore, portMAX_DELAY))
#define COMM_END xSemaphoreGive(comm_semaphore)

uint8_t comm_init(void)
{
	tlvbuf[0] = 0x0A;
	tlvbuf[1] = 0x42;
	comm_semaphore = xSemaphoreCreateBinary();
	if (comm_semaphore == NULL)
		return 0;
	xSemaphoreGive(comm_semaphore);
	rx_queue = xQueueCreate(COMM_RX_QUEUE_LENGTH, sizeof(uint8_t));
	tx_queue = xQueueCreate(COMM_TX_QUEUE_LENGTH, sizeof(uint8_t));
	USB_VCP_Init(rx_queue, tx_queue);
	return 0;
}

/* - READ */

uint8_t comm_read(void)
{
	uint8_t c;

	while (xQueueReceive(rx_queue, &c, portMAX_DELAY) != pdPASS)
		;
	return c;
}

/* - WRITE */

static inline void VCP_Transmit_Bytes(const uint8_t *_buf, const int _len)
{
	int i;

#ifndef COMM_BLOCK_ON_DISCONNECT
	if (!USB_VCP_Connected()) {
		xQueueReset(tx_queue);
		return;
	}
#endif /* COMM_BLOCK_ON_DISCONNECT */
	for (i = 0; i < _len; i++) {
#if !defined(COMM_TX_TIMEOUT) || COMM_TX_TIMEOUT < 0
		while (pdPASS != xQueueSend(tx_queue, &_buf[i], portMAX_DELAY))
			;
#else /* !defined(COMM_TX_TIMEOUT) || COMM_TX_TIMEOUT < 0 */
		if (xQueueSend(tx_queue, &_buf[i],
				COMM_TX_TIMEOUT / portTICK_PERIOD_MS) != pdPASS)
			return;
#endif /* !defined(COMM_TX_TIMEOUT) || COMM_TX_TIMEOUT < 0 */
	}
}

void comm_begin_packet(const size_t length, const enum comm_type type)
{
	COMM_BEGIN;
	tlvbuf[2] = (uint8_t)type;
	tlvbuf[3] = (length & 0xFF00) >> 8;
	tlvbuf[4] = length & 0x00FF;
	VCP_Transmit_Bytes(tlvbuf, 5);
}

void comm_end_packet(void)
{
	COMM_END;
}

void comm_send(const void *data, const size_t length)
{
	VCP_Transmit_Bytes(data, length);
}

void comm_send_packet(
	const void *data, const size_t length, const enum comm_type type)
{
	comm_begin_packet(length, type);
	VCP_Transmit_Bytes(data, length);
	comm_end_packet();
}

void comm_write_raw(const void *data, const size_t length)
{
	comm_send_packet(data, length, COMM_TYPE_MESSAGE);
}

void comm_write(const char *string)
{
	comm_send_packet(string, strlen(string), COMM_TYPE_MESSAGE);
}

uint8_t comm_output_waiting(void)
{
	return (tx_queue != NULL && uxQueueMessagesWaiting(tx_queue));
}

#endif /* INCLUDE_BOARD_LIB */
