#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include <FreeRTOS.h>
#include <queue.h>
#include <task.h>
#include <semphr.h>

#include <upcn.h>
#include <contactManager.h>
#include <routerTask.h>
#include <groundStation.h>
#include <groundStationTask.h>
#include <beaconProcessor.h>
#include <beaconGenerator.h>

struct contact_manager_task_parameters {
	SemaphoreHandle_t semaphore;
	QueueHandle_t control_queue;
	QueueHandle_t beacon_queue;
	QueueHandle_t router_signaling_queue;
	struct contact_list **contact_list_ptr;
};

struct contact_info {
	struct contact *contact;
	uint16_t expected_beacons;
	uint16_t received_beacons;
};

static TaskHandle_t gs_tasks[MAX_CONCURRENT_CONTACTS];
static QueueHandle_t gs_queues[MAX_CONCURRENT_CONTACTS];

static struct contact_info current_contacts[MAX_CONCURRENT_CONTACTS];
static int8_t current_contact_count;
static uint64_t next_contact_time = UINT64_MAX;

static bool contact_active(const struct contact *contact)
{
	int8_t i;

	for (i = 0; i < current_contact_count; i++) {
		if (current_contacts[i].contact == contact)
			return true;
	}
	return false;
}

static void reset_bundles(int8_t index)
{
	struct contact *c = current_contacts[index].contact;
	struct ground_station_task_command cmd;
	struct routed_bundle_list **tmp;
	uint8_t fin = 0;

	while (pdTRUE == xQueueReceive(gs_queues[index], &cmd, 0)) {
		if (cmd.type != GS_COMMAND_BUNDLES) {
			if (cmd.type == GS_COMMAND_FINALIZE)
				fin = 1;
			continue;
		}
		if (c->contact_bundles == NULL) {
			c->contact_bundles = cmd.bundles;
		} else {
			tmp = &cmd.bundles;
			while (*tmp != NULL)
				tmp = &(*tmp)->next;
			*tmp = c->contact_bundles;
			c->contact_bundles = cmd.bundles;
		}
	}
	if (fin) /* NOTE: Should never occur */
		ground_station_task_delete(gs_queues[index]);
}

static int8_t remove_expired_contacts(
	struct contact_list *contact_list, const uint64_t current_timestamp,
	struct contact_info list[])
{
	/* Check for ending contacts */
	/* We do this first as it frees space in the list */
	int8_t i, c, removed = 0;
	QueueHandle_t tmpq;
	TaskHandle_t tmpt;

	for (i = current_contact_count - 1; i >= 0; i--) {
		if (current_contacts[i].contact->to <= current_timestamp) {
			/* Unset "active" constraint */
			current_contacts[i].contact->active = 0;
			/* Move bundles to contact again for rescheduling */
			reset_bundles(i);
			list[removed++] = current_contacts[i];
			/* If it's not the last element, we have to move mem */
			if (i != current_contact_count - 1) {
				for (c = i; c < current_contact_count; c++) {
					current_contacts[c] =
						current_contacts[c + 1];
					/* Swap queue and task as well */
					tmpq = gs_queues[c];
					gs_queues[c] = gs_queues[c + 1];
					gs_queues[c + 1] = tmpq;
					tmpt = gs_tasks[c];
					gs_tasks[c] = gs_tasks[c + 1];
					gs_tasks[c + 1] = tmpt;
				}
			}
			current_contact_count--;
		}
	}
	return removed;
}

static uint8_t check_upcoming(
	struct contact *c, struct contact_info list[], const uint8_t index)
{
	uint16_t exp;

	if (!contact_active(c)
		&& current_contact_count < MAX_CONCURRENT_CONTACTS
	) {
		/* Set "active" constraint */
		c->active = 1;
		/* Add contact */
		current_contacts[current_contact_count].contact = c;
		if (HAS_FLAG(c->ground_station->flags, GS_FLAG_DISCOVERED)) {
			exp = (c->to - c->from) / c->ground_station
				->rrnd_info->beacon_period;
		} else {
			exp = 0;
		}
		current_contacts[current_contact_count].expected_beacons = exp;
		current_contacts[current_contact_count].received_beacons = 0;
		list[index] = current_contacts[current_contact_count];
		current_contact_count++;
		return 1;
	}
	return 0;
}

static int8_t process_upcoming_list(
	struct contact_list *contact_list, const uint64_t current_timestamp,
	struct contact_info list[])
{
	int8_t added = 0;
	struct contact_list *cur_entry;

	next_contact_time = UINT64_MAX;
	cur_entry = contact_list;
	while (cur_entry != NULL) {
		if (cur_entry->data->from <= current_timestamp) {
			if (cur_entry->data->to > current_timestamp) {
				added += check_upcoming(
					cur_entry->data, list, added);
				if (cur_entry->data->to < next_contact_time)
					next_contact_time = cur_entry->data->to;
			}
		} else {
			if (cur_entry->data->from < next_contact_time)
				next_contact_time = cur_entry->data->from;
			/* As our contact_list is sorted ascending by */
			/* from-time we can stop checking here */
			break;
		}
		cur_entry = cur_entry->next;
	}
	return added;
}

static void hand_over_contact_bundles(int8_t index, struct contact_info c)
{
	struct ground_station_task_command command = {
		.type = GS_COMMAND_BUNDLES,
		.bundles = NULL,
		.contact = c.contact
	};

	ASSERT(c.contact != NULL);
	if (c.contact->contact_bundles == NULL)
		return;
	command.bundles = c.contact->contact_bundles;
	c.contact->contact_bundles = NULL;
	while (pdFALSE == xQueueSendToBack(
			gs_queues[index], &command, portMAX_DELAY))
		;
}

static void process_current_contacts(void)
{
	int8_t i;

	for (i = 0; i < current_contact_count; i++)
		hand_over_contact_bundles(i, current_contacts[i]);
}

static uint8_t check_for_contacts(struct contact_list *contact_list,
	struct contact_info removed_contacts[])
{
	int8_t i;
	static struct contact_info added_contacts[MAX_CONCURRENT_CONTACTS];
	uint64_t current_timestamp = clock_get_timestamp();
	int8_t removed_count = remove_expired_contacts(
		contact_list, current_timestamp, removed_contacts);
	int8_t added_count = process_upcoming_list(
		contact_list, current_timestamp, added_contacts);

	for (i = 0; i < added_count; i++) {
		LOGI("ContactManager: Contact added",
			added_contacts[i].contact);
		hand_over_contact_bundles(i, added_contacts[i]);
	}
	for (i = 0; i < removed_count; i++) {
		LOGI("ContactManager: Contact removed",
			removed_contacts[i].contact);
	}
	return removed_count;
}

static void inform_router(
	enum router_signal_type type, void *data, QueueHandle_t queue)
{
	struct router_signal signal = {
		.type = type,
		.data = data
	};

	ASSERT(data != NULL);
	while (pdFALSE == xQueueSendToBack(
			queue, &signal, portMAX_DELAY))
		;
}

/* In this function we can assume that contact_list will not change */
static void manage_contacts(
	struct contact_list *contact_list, enum contact_manager_signal signal,
	SemaphoreHandle_t semphr, QueueHandle_t queue)
{
	static struct contact_info rem[MAX_CONCURRENT_CONTACTS];
	int8_t removed, i;

	ASSERT(semphr != NULL);
	ASSERT(queue != NULL);
	while (pdFALSE == xSemaphoreTake(semphr, portMAX_DELAY))
		;
	if (HAS_FLAG(signal, CM_SIGNAL_BUNDLE_SCHEDULED))
		process_current_contacts();
	/* If there only was a new bundle, skip further checks */
	if (signal == CM_SIGNAL_BUNDLE_SCHEDULED) {
		xSemaphoreGive(semphr);
		return;
	}
	removed = check_for_contacts(contact_list, rem);
	xSemaphoreGive(semphr);
	for (i = 0; i < removed; i++) {
		/* The contact has to be deleted first... */
		inform_router(ROUTER_SIGNAL_CONTACT_OVER,
			rem[i].contact, queue);
	}
}

static void start_gs_tasks(QueueHandle_t router_sig_queue)
{
	int8_t i;
	struct ground_station_task_creation_result tmp;

	for (i = 0; i < MAX_CONCURRENT_CONTACTS; i++) {
		tmp = ground_station_task_create(router_sig_queue);
		ASSERT(tmp.result == UPCN_OK);
		gs_tasks[i] = tmp.task_handle;
		gs_queues[i] = tmp.queue_handle;
	}
}

static void contact_manager_task(void *cm_parameters)
{
	struct contact_manager_task_parameters *parameters
		= (struct contact_manager_task_parameters *)cm_parameters;
	int8_t led_state = 0;
	enum contact_manager_signal signal = CM_SIGNAL_NONE;
	struct beacon *beacon;
	uint64_t cur_time, next_time, delay;

	ASSERT(parameters != NULL);
	start_gs_tasks(parameters->router_signaling_queue);
	LOG("ContactManager: Started up successfully");
	for (;;) {
		led_set((led_state = 1 - led_state) + 3);
		while (pdTRUE == xQueueReceive(
			parameters->beacon_queue, &beacon, 0)
		) {
			/* TODO: Do sth. with the result! */
			beacon_processor_process(beacon,
				parameters->router_signaling_queue);
		}
		/* TODO: Only begin C. with indic. of bidirectionalty */
		if (signal != CM_SIGNAL_NONE) { /* TODO: end by query to BP */
			manage_contacts(*parameters->contact_list_ptr, signal,
				parameters->semaphore,
				parameters->router_signaling_queue);
		}
		signal = CM_SIGNAL_UNKNOWN;
		cur_time = clock_get_timestamp_ms();
#ifdef ND_DISABLE_BEACONS
		next_time = UINT64_MAX;
#else /* ND_DISABLE_BEACONS */
		next_time = beacon_generator_check_send(cur_time);
#endif /* ND_DISABLE_BEACONS */
		next_time = MIN(next_time, next_contact_time * 1000);
		if (next_time > (cur_time + CONTACT_CHECKING_MAX_PERIOD))
			delay = CONTACT_CHECKING_MAX_PERIOD;
		else if (next_time <= cur_time)
			continue;
		else
			delay = next_time - cur_time;
		/*LOGA("ContactManager: Suspended on queue",*/
		/*	(uint8_t)(delay / 1000), LOG_NO_ITEM);*/
		xQueueReceive(parameters->control_queue, &signal,
			delay / portTICK_PERIOD_MS + 1);
	}
}

struct contact_manager_params contact_manager_start(
	QueueHandle_t router_signaling_queue, struct contact_list **clistptr)
{
	struct contact_manager_params ret = {
		.semaphore = NULL,
		.control_queue = NULL,
		.beacon_queue = NULL
	};
	TaskHandle_t cm_task;
	QueueHandle_t queue, beacon_queue;
	struct contact_manager_task_parameters *cmt_params;
	SemaphoreHandle_t semaphore = xSemaphoreCreateMutex();

	if (semaphore == NULL)
		return ret;
	xSemaphoreGive(semaphore);
	queue = xQueueCreate(1, sizeof(enum contact_manager_signal));
	beacon_queue = xQueueCreate(5, sizeof(void *));
	if (queue == NULL) {
		vSemaphoreDelete(semaphore);
		return ret;
	}
	cmt_params = malloc(sizeof(struct contact_manager_task_parameters));
	if (cmt_params == NULL) {
		vSemaphoreDelete(semaphore);
		vQueueDelete(queue);
		return ret;
	}
	cmt_params->semaphore = semaphore;
	cmt_params->control_queue = queue;
	cmt_params->beacon_queue = beacon_queue;
	cmt_params->router_signaling_queue = router_signaling_queue;
	cmt_params->contact_list_ptr = clistptr;
	xTaskCreate(contact_manager_task, "contact_manager_task",
		CONTACT_MANAGER_TASK_STACK_SIZE, cmt_params,
		CONTACT_MANAGER_TASK_PRIORITY, &cm_task);
	vTaskSetApplicationTaskTag(cm_task, (void *)CONTACT_MANAGER_TASK_TAG);
	ret.semaphore = semaphore;
	ret.control_queue = queue;
	ret.beacon_queue = beacon_queue;
	return ret;
}

uint64_t contact_manager_get_next_contact_time(void)
{
	return next_contact_time;
}

uint8_t contact_manager_in_contact(void)
{
	return current_contact_count != 0;
}

void contact_manager_reset_time(void)
{
	next_contact_time = clock_get_timestamp();
}
