#include <stdlib.h>
#include <string.h>

#include <upcn.h>
#include <bundle.h>
#include <bundleStorageManager.h>
#include <custodyManager.h>

static struct bundle *accepted_bundles[CUSTODY_MAX_BUNDLE_COUNT];
static int accepted_bundle_count;

static int get_index(struct bundle *bundle)
{
	int i;

	for (i = 0; i < accepted_bundle_count; i++) {
		if (accepted_bundles[i] == bundle)
			return i;
	}
	return -1;
}

static int find(
	uint64_t creation_timestamp, uint16_t sequence_number,
	uint16_t source_eid_length, char *source_eid,
	uint32_t fragment_offset, uint32_t fragment_length)
{
	int i;
	struct bundle *b;

	for (i = 0; i < accepted_bundle_count; i++) {
		b = accepted_bundles[i];
		if (
			b->creation_timestamp == creation_timestamp
			&& b->sequence_number == sequence_number
			&& bundle_eid_equals_string(b, b->source_eid,
				source_eid, source_eid_length)
			&& (!HAS_FLAG(b->proc_flags, BUNDLE_FLAG_IS_FRAGMENT)
				|| (b->fragment_offset == fragment_offset
				&& b->payload_block->length == fragment_length))
		) {
			return i;
		}
	}
	return -1;
}

int custody_manager_has_redundant_bundle(struct bundle *bundle)
{
	char *eid;
	int index;

	eid = bundle_read_eid(bundle, bundle->source_eid);
	index = find(
		bundle->creation_timestamp, bundle->sequence_number,
		strlen(eid), eid,
		bundle->fragment_offset, bundle->payload_block->length);
	free(eid);
	return index;
}

int custody_manager_storage_is_acceptable(struct bundle *bundle)
{
	return !(
		accepted_bundle_count >= CUSTODY_MAX_BUNDLE_COUNT
		|| get_index(bundle) != -1
		|| bundle_get_serialized_size(bundle) > CUSTODY_MAX_BUNDLE_SIZE
		|| !HAS_FLAG(bundle->proc_flags, BUNDLE_FLAG_SINGLETON_ENDPOINT)
	);
}

int custody_manager_has_accepted(struct bundle *bundle)
{
	return (get_index(bundle) != -1);
}

struct bundle *custody_manager_get_by_record(
	struct bundle_administrative_record *record)
{
	int index;

	index = find(
		record->bundle_creation_timestamp,
		record->bundle_sequence_number,
		record->bundle_source_eid_length,
		record->bundle_source_eid,
		record->fragment_offset,
		record->fragment_length
	);
	if (index == -1)
		return NULL;
	else
		return accepted_bundles[index];
}

/* 5.10.1 */
enum upcn_result custody_manager_accept(struct bundle *bundle)
{
	static const char own_scheme[] = UPCN_SCHEME;
	static const char own_ssp[] = UPCN_SSP;
	uint16_t old_length;

	/* Should be checked by bundle processor */
	ASSERT(!custody_manager_has_redundant_bundle(bundle));
	ASSERT(custody_manager_storage_is_acceptable(bundle));
	/* Add to list */
	accepted_bundles[accepted_bundle_count++] = bundle;
	/* Add ret. constraint */
	bundle->ret_constraints |= BUNDLE_RET_CONSTRAINT_CUSTODY_ACCEPTED;
	/* Add own EID as custodian and to dict */
	old_length = bundle->dict_length;
	bundle->dict_length += sizeof(own_scheme) + sizeof(own_ssp);
	bundle->dict = realloc(bundle->dict, bundle->dict_length);
	if (bundle->dict == NULL)
		return UPCN_FAIL;
	bundle->custodian_eid.scheme_offset = old_length;
	memcpy(bundle->dict + old_length, own_scheme, sizeof(own_scheme));
	old_length += sizeof(own_scheme);
	bundle->custodian_eid.ssp_offset = old_length;
	memcpy(bundle->dict + old_length, own_ssp, sizeof(own_ssp));
	bundle_recalculate_header_length(bundle);
	return UPCN_OK;
}

/* 5.10.2 */
void custody_manager_release(struct bundle *bundle)
{
	size_t num;
	int index = get_index(bundle);

	if (index == -1)
		return;
	if (index != (accepted_bundle_count - 1)) {
		num = accepted_bundle_count - index - 1;
		memmove(
			accepted_bundles + index,
			accepted_bundles + index + 1,
			num
		);
	}
	accepted_bundle_count--;
	bundle->ret_constraints &= ~BUNDLE_RET_CONSTRAINT_CUSTODY_ACCEPTED;
	if (bundle->ret_constraints == BUNDLE_RET_CONSTRAINT_NONE) {
		bundle_storage_delete(bundle->id);
		bundle_drop(bundle);
	}
}
