#include <stdlib.h>

#include <FreeRTOS.h>
#include <queue.h>
#include <task.h>

#include <upcn.h>
#include <bundle.h>
#include <groundStation.h>
#include <bundleSerializer.h>
#include <routerTask.h>
#include <bundleStorageManager.h>
#include <groundStationTask.h>

struct gs_task_parameters {
	QueueHandle_t queue_handle;
	QueueHandle_t router_signaling_queue;
};

static inline void report_bundle(QueueHandle_t signaling_queue,
	struct routed_bundle *bundle, enum router_signal_type type)
{
	struct router_signal signal = {
		.type = type,
		.data = bundle
	};

	while (pdFALSE == xQueueSendToBack(
			signaling_queue, &signal, portMAX_DELAY))
		;
}

static void ground_station_task(void *p)
{
	QueueHandle_t cmd_queue =
		((struct gs_task_parameters *)p)->queue_handle;
	QueueHandle_t rt_queue =
		((struct gs_task_parameters *)p)->router_signaling_queue;
	struct ground_station_task_command cmd;
	struct bundle *b;
	struct routed_bundle_list *cur;
	enum upcn_result s;

	LOGI("GroundStationTask: Started successfully",
		((struct gs_task_parameters *)p)->queue_handle);
	for (;;) {
		if (!xQueueReceive(cmd_queue, &cmd, portMAX_DELAY))
			continue;
		else if (cmd.type == GS_COMMAND_FINALIZE)
			break;
		else if (cmd.type != GS_COMMAND_BUNDLES)
			continue;
		/* GS_COMMAND_BUNDLES received */
		while (cmd.bundles != NULL) {
			cur = cmd.bundles;
			cmd.bundles = cmd.bundles->next;
			cur->data->serialized++;
			b = bundle_storage_get(cur->data->id);
			/* LOGI("GroundStationTask: Sending bundle", b->id); */
#ifdef VERBOSE
#ifdef BUNDLE_PRINT
			bundle_print_stack_conserving(b);
#endif /* BUNDLE_PRINT */
#endif /* VERBOSE */
			comm_lock();
			comm_begin_packet(
				bundle_get_serialized_size(b),
				COMM_TYPE_BUNDLE);
			s = bundle_serialize(b, &comm_send);
			comm_end_packet();
			comm_unlock();
			if (s == UPCN_OK) {
				cur->data->transmitted++;
				report_bundle(rt_queue, cur->data,
					ROUTER_SIGNAL_TRANSMISSION_SUCCESS);
			} else {
				report_bundle(rt_queue, cur->data,
					ROUTER_SIGNAL_TRANSMISSION_FAILURE);
			}
			/* Free only the RB list, the RB is reported */
			free(cur);
		}
	}
	/* Free vars and delete task */
	free(p);
	vQueueDelete(cmd_queue);
	vTaskDelete(NULL);
}

struct ground_station_task_creation_result ground_station_task_create(
	QueueHandle_t router_signaling_queue)
{
	static uint8_t ctr = 1;
	static char tname_buf[6];
	struct gs_task_parameters *param;
	struct ground_station_task_creation_result res = {
		.task_handle = NULL,
		.queue_handle = NULL,
		.result = UPCN_FAIL
	};

	res.queue_handle = xQueueCreate(GROUND_STATION_TASK_QUEUE_LENGTH,
		sizeof(struct ground_station_task_command));
	if (res.queue_handle == NULL)
		return res;
	param = malloc(sizeof(struct gs_task_parameters));
	if (param == NULL) {
		vQueueDelete(res.queue_handle);
		return res;
	}
	param->queue_handle = res.queue_handle;
	param->router_signaling_queue = router_signaling_queue;
	tname_buf[0] = 'g';
	tname_buf[1] = 's';
	sprintu32(tname_buf + 2, ctr++);
	if (xTaskCreate(ground_station_task, tname_buf,
		GROUND_STATION_TASK_STACK_SIZE, param,
		GROUND_STATION_TASK_PRIORITY, &res.task_handle) == pdPASS
	) {
		vTaskSetApplicationTaskTag(res.task_handle,
			(void *)GROUND_STATION_TASK_TAG);
		res.result = UPCN_OK;
	} else {
		vQueueDelete(res.queue_handle);
		free(param);
	}
	return res;
}

void ground_station_task_delete(QueueHandle_t queue)
{
	struct ground_station_task_command command = {
		.type = GS_COMMAND_FINALIZE,
		.bundles = NULL,
		.contact = NULL
	};

	ASSERT(queue != NULL);
	while (pdFALSE == xQueueSendToBack(queue, &command, portMAX_DELAY))
		;
}
