#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#include <stm32f4xx.h>

#include <FreeRTOS.h>
#include <task.h>
#include <queue.h>

#include <upcn.h>
#include <routerTask.h>
#include <bundleParser.h>
#include <inputParser.h>
#include <routerParser.h>
#include <bundleProcessor.h>
#include <bundleStorageManager.h>
#include <beacon.h>
#include <beaconParser.h>
#include <beaconProcessor.h>
#include <eidManager.h>

static QueueHandle_t router_signaling_queue;
static QueueHandle_t bundle_signaling_queue;

static struct parser *cur_parser;

static struct input_parser input_parser;
static struct router_parser router_parser;
static struct bundle_parser bundle_parser;
static struct beacon_parser beacon_parser;

static void reset_parser(void)
{
	input_parser_reset(&input_parser);
	cur_parser = input_parser.basedata;
	ASSERT(router_parser_reset(&router_parser) == UPCN_OK);
	ASSERT(bundle_parser_reset(&bundle_parser) == UPCN_OK);
	ASSERT(beacon_parser_reset(&beacon_parser) == UPCN_OK);
}

static void forward_to_specific_parser(uint8_t byte)
{
	if (input_parser.stage != INPUT_EXPECT_DATA) {
		/* Current parser is input_parser */
		input_parser_read_byte(&input_parser, byte);
	} else {
		switch (input_parser.type) {
		case INPUT_TYPE_ROUTER_COMMAND_DATA:
			cur_parser = router_parser.basedata;
			router_parser_read_byte(&router_parser, byte);
			break;
		case INPUT_TYPE_BUNDLE_DATA:
			cur_parser = bundle_parser.basedata;
			bundle_parser_read_byte(&bundle_parser, byte);
			break;
		case INPUT_TYPE_BEACON_DATA:
			cur_parser = beacon_parser.basedata;
			beacon_parser_read_byte(&beacon_parser, byte);
			break;
		default:
			reset_parser();
			return;
		}
	}
}

static void router_command_send(struct router_command *cmd)
{
	struct router_signal signal = {
		.type = ROUTER_SIGNAL_PROCESS_COMMAND,
		.data = cmd
	};

	ASSERT(cmd != NULL);
	while (pdFALSE == xQueueSendToBack(
			router_signaling_queue, &signal, portMAX_DELAY))
		;
}

static void bundle_send(struct bundle *bundle)
{
	bundleid_t new_id;

	ASSERT(bundle != NULL);
	new_id = bundle_storage_add(bundle);
	if (new_id != BUNDLE_INVALID_ID)
		bundle_processor_inform(
			bundle_signaling_queue,
			new_id,
			BP_SIGNAL_BUNDLE_INCOMING,
			BUNDLE_SR_REASON_NO_INFO);
	else
		bundle_free(bundle);
}

static void beacon_send(struct beacon *beacon)
{
	struct router_signal signal = {
		.type = ROUTER_SIGNAL_PROCESS_BEACON,
		.data = beacon
	};

	ASSERT(beacon != NULL);
	while (pdFALSE == xQueueSendToBack(
			router_signaling_queue, &signal, portMAX_DELAY))
		;
}

static void input_task(void *args)
{
	ASSERT(cur_parser = input_parser_init(&input_parser));
	ASSERT(router_parser_init(&router_parser, &router_command_send));
	ASSERT(bundle_parser_init(&bundle_parser, &bundle_send));
	ASSERT(beacon_parser_init(&beacon_parser, &beacon_send));

	for (;;) {
		forward_to_specific_parser(comm_read());
		if (cur_parser->status != PARSER_STATUS_GOOD)
			reset_parser();
	}
}

static void init_subsystems(void)
{
	uint8_t i, randlen = (BEACON_PRIVATE_KEY_LENGTH / 4)
		+ ((BEACON_PRIVATE_KEY_LENGTH % 4) ? 1 : 0);
	uint32_t *randbuf = malloc(randlen * 4);

	ASSERT(randbuf != NULL);
	eidmanager_init();
	for (i = 0; i < randlen; i++)
		randbuf[i] = random_get();
	ASSERT(beacon_processor_init((uint8_t *)randbuf) == UPCN_OK);
	free(randbuf);
}

void init(void)
{
	firmware_init();
	led_set(0);
	init_subsystems();
	comm_printf("\n\nThis is uPCN, compiled %s %s\n", __DATE__, __TIME__);
	comm_printf("All subsystems initialized, starting RTOS...\n\n");
}

#ifdef TEST_SGP4_PERFORMANCE
void sgp4testtask(void *task_params);
#endif /* TEST_SGP4_PERFORMANCE */

void start_tasks(void)
{
	TaskHandle_t cur_task;

	/* Initialize queues to communicate with the subsystems */
	router_signaling_queue
		= xQueueCreate(ROUTER_QUEUE_LENGTH,
			sizeof(struct router_signal));
	ASSERT(router_signaling_queue != NULL);
	bundle_signaling_queue
		= xQueueCreate(BUNDLE_QUEUE_LENGTH,
			sizeof(struct bundle_processor_signal));
	ASSERT(bundle_signaling_queue != NULL);

	struct router_task_parameters *router_task_params =
		malloc(sizeof(struct router_task_parameters));
	ASSERT(router_task_params != NULL);
	router_task_params->router_signaling_queue = router_signaling_queue;
	router_task_params->bundle_processor_signaling_queue
		= bundle_signaling_queue;

	struct bundle_processor_task_parameters *bundle_processor_task_params
		= malloc(sizeof(struct bundle_processor_task_parameters));
	ASSERT(bundle_processor_task_params != NULL);
	bundle_processor_task_params->router_signaling_queue
		= router_signaling_queue;
	bundle_processor_task_params->signaling_queue = bundle_signaling_queue;

	xTaskCreate(router_task, "router_task",
		DEFAULT_TASK_STACK_SIZE, router_task_params,
		ROUTER_TASK_PRIORITY, &cur_task);
	vTaskSetApplicationTaskTag(cur_task, (void *)ROUTER_TASK_TAG);
	xTaskCreate(bundle_processor_task, "bundle_processor_task",
		DEFAULT_TASK_STACK_SIZE, bundle_processor_task_params,
		BUNDLE_PROCESSOR_TASK_PRIORITY, &cur_task);
	vTaskSetApplicationTaskTag(cur_task, (void *)BUNDLE_PROCESSOR_TASK_TAG);
	xTaskCreate(input_task, "input_task",
		INPUT_TASK_STACK_SIZE, NULL, INPUT_TASK_PRIORITY, &cur_task);
	vTaskSetApplicationTaskTag(cur_task, (void *)INPUT_TASK_TAG);
#ifdef TEST_SGP4_PERFORMANCE
	xTaskCreate(sgp4testtask, "sgp4test",
		4096, NULL, CONTACT_MANAGER_TASK_PRIORITY, &cur_task);
	vTaskSetApplicationTaskTag(cur_task, (void *)13);
#endif /* TEST_SGP4_PERFORMANCE */
}

int start_os(void)
{
	vTaskStartScheduler();
	/* Should never get here! */
	NVIC_SystemReset();
	return 0;
}
