#include <stdlib.h>

#include <upcn.h>
#include <comm.h>
#include <inputParser.h>
#include <contactManager.h>
#include <groundStation.h>
#include <routingTable.h>
#include <rrnd.h>
#include <satpos.h>
#include <mini-printf.h>
#include <beaconProcessor.h>

static const uint8_t DATA_ESCAPE_DELIMITER; /* 0x00 */
static const uint8_t DATA_BEGIN_MARKER       = 0xFF;
static const uint8_t DATA_END_MARKER         = 0xFF;
static const uint8_t TYPE_DELIMITER          = ':';

static const uint8_t ECHO_MESSAGE[]        =
	"This is uPCN, compiled " __DATE__ " " __TIME__ ". " \
	"Greetings to you, earthling.";

static const unsigned int RRND_CHARSTATE_SUCCESS = RRND_STATUS_SUCCESS;
static const unsigned int RRND_CHARSTATE_FAILURE = RRND_STATUS_FAILED;

static void print_gs(char *eid)
{
	static char gsbuf[4096];
	struct ground_station *gs = routing_table_lookup_ground_station(eid);
	size_t buflen;

	buflen = rrnd_print_gs_info(
		gsbuf, 4096, eid, gs ? gs->rrnd_info : NULL);
	comm_send_packet(gsbuf, buflen, COMM_TYPE_GS_INFO);
}

static void infer_ct(char *eid)
{
	uint64_t cur_time = clock_get_timestamp_ms();
	struct ground_station *gs = routing_table_lookup_ground_station(eid);
	enum rrnd_status res;

	if (gs == NULL || gs->rrnd_info == NULL) {
		LOG("GS for inferring contact not found");
		comm_send_packet(&RRND_CHARSTATE_FAILURE,
			RRND_STATUS_SIZEOF, COMM_TYPE_RRND_STATUS);
		return;
	}
	res = rrnd_infer_contact(gs->rrnd_info, cur_time, satpos_get,
		satpos_get_age(cur_time));
#ifdef INTEGRATE_DEMANDED_CONTACTS
	if (!HAS_FLAG(res, RRND_STATUS_FAILED))
		routing_table_integrate_inferred_contact(gs, 0);
#endif /* INTEGRATE_DEMANDED_CONTACTS */
	comm_send_packet(&res, RRND_STATUS_SIZEOF, COMM_TYPE_RRND_STATUS);
}

void input_parser_read_byte(struct input_parser *parser, uint8_t byte)
{
	switch (parser->stage) {
	case INPUT_EXPECT_DATA_ESCAPE_DELIMITER:
		if (byte == DATA_ESCAPE_DELIMITER)
			parser->stage = INPUT_EXPECT_DATA_BEGIN_MARKER;
		else if (byte == DATA_BEGIN_MARKER)
			parser->stage = INPUT_EXPECT_TYPE;
		/*
		 * We do not need an Input Parser reset here as parsing
		 * has not begun and we can simply wait for the "next"
		 * escape delimiter...
		 */
		break;
	case INPUT_EXPECT_DATA_BEGIN_MARKER:
		if (byte == DATA_BEGIN_MARKER)
			parser->stage = INPUT_EXPECT_TYPE;
		else if (byte != DATA_ESCAPE_DELIMITER)
			parser->stage = INPUT_EXPECT_DATA_ESCAPE_DELIMITER;
		/* See above... */
		break;
	case INPUT_EXPECT_TYPE:
		parser->stage = INPUT_EXPECT_TYPE_DELIMITER;
		if (byte >= INPUT_TYPE_ROUTER_COMMAND_DATA
				&& byte <= INPUT_TYPE_MAX)
			parser->type = (enum input_type)byte;
		else if (byte == DATA_ESCAPE_DELIMITER)
			parser->stage = INPUT_EXPECT_DATA_BEGIN_MARKER;
		else if (byte != DATA_BEGIN_MARKER)
			parser->basedata->status = PARSER_STATUS_ERROR;
		break;
	case INPUT_EXPECT_TYPE_DELIMITER:
		if (byte == TYPE_DELIMITER) {
			parser->intdata = 0;
			parser->intdata_index = 0;
			if (parser->type == INPUT_TYPE_SET_TIME)
				parser->stage = INPUT_EXPECT_TIME;
			else if (parser->type >= INPUT_TYPE_GET_GS &&
					parser->type <=
					INPUT_TYPE_RRND_INFER_CONTACT)
				parser->stage = INPUT_EXPECT_STRING_LENGTH;
			else if (parser->type >= INPUT_TYPE_DBG_QUERY)
				parser->stage = INPUT_EXPECT_DATA_END_MARKER;
			else
				parser->stage = INPUT_EXPECT_DATA;
		} else {
			parser->basedata->status = PARSER_STATUS_ERROR;
		}
		break;
	case INPUT_EXPECT_TIME:
		parser->intdata <<= 8;
		parser->intdata |= byte;
		if (++parser->intdata_index == 4)
			parser->stage = INPUT_EXPECT_DATA_END_MARKER;
		break;
	case INPUT_EXPECT_STRING_LENGTH:
		parser->intdata <<= 8;
		parser->intdata |= byte;
		if (++parser->intdata_index == 2) {
			if (parser->intdata <= 1024) {
				parser->string = malloc(parser->intdata + 1);
				if (parser->string != NULL) {
					parser->stage = INPUT_EXPECT_STRING;
					parser->intdata_index = 0;
					break;
				}
			}
			parser->basedata->status = PARSER_STATUS_ERROR;
		}
		break;
	case INPUT_EXPECT_STRING:
		parser->string[parser->intdata_index] = (char)byte;
		parser->intdata_index++;
		parser->intdata--;
		if (parser->intdata == 0) {
			parser->string[parser->intdata_index] = '\0';
			parser->stage = INPUT_EXPECT_DATA_END_MARKER;
		}
		break;
	case INPUT_EXPECT_DATA_END_MARKER:
		if (byte == DATA_END_MARKER) {
			parser->basedata->status = PARSER_STATUS_DONE;
			switch (parser->type) {
			case INPUT_TYPE_RESET_TIME:
				clock_reset();
				contact_manager_reset_time();
				break;
			case INPUT_TYPE_SET_TIME:
				clock_set_timestamp(parser->intdata);
				contact_manager_reset_time();
				LOGI("Time set", parser->intdata);
				comm_send_packet("\x01", 1,
					COMM_TYPE_GENERIC_RESULT);
				break;
			case INPUT_TYPE_DBG_QUERY:
#ifdef UPCN_TEST_BUILD
				upcntest_print();
#endif /* UPCN_TEST_BUILD */
				upcn_print_info();
				upcn_dbg_memprint();
				upcn_dbg_memstat_print();
				upcn_dbg_printtrace();
				upcn_dbg_printtraces();
				upcn_dbg_printlogs();
				break;
			case INPUT_TYPE_RESET_STATS:
				upcn_dbg_resettrace();
				upcn_dbg_memstat_reset();
				comm_send_packet("\x01", 1,
					COMM_TYPE_GENERIC_RESULT);
				LOG("Trace stats reset");
				break;
			case INPUT_TYPE_CLEAR_TRACES:
				upcn_dbg_cleartraces();
				comm_send_packet("\x01", 1,
					COMM_TYPE_GENERIC_RESULT);
				LOG("Trace storage reset");
				break;
			case INPUT_TYPE_STORE_TRACE:
				upcn_dbg_storetrace();
				comm_send_packet("\x01", 1,
					COMM_TYPE_GENERIC_RESULT);
				LOG("Trace stored");
				break;
			case INPUT_TYPE_ECHO:
				comm_send_packet(
					ECHO_MESSAGE, sizeof(ECHO_MESSAGE) - 1,
					COMM_TYPE_ECHO);
				break;
			case INPUT_TYPE_RESET:
				LOG("SYSTEM RESET");
				comm_send_packet("\x01", 1,
					COMM_TYPE_GENERIC_RESULT);
				upcn_reset_system();
				break;
			case INPUT_TYPE_GET_GS:
				print_gs(parser->string);
				break;
			case INPUT_TYPE_SET_SATPOS_TLE:
				satpos_init(parser->string,
					clock_get_timestamp_ms());
				if (satpos_last_result() == SATPOS_SUCCESS) {
					LOG("SATPOS initialized successfully");
					comm_send_packet(
						&RRND_CHARSTATE_SUCCESS,
						RRND_STATUS_SIZEOF,
						COMM_TYPE_RRND_STATUS);
				} else {
					LOG("SATPOS initialization failed");
					comm_send_packet(
						&RRND_CHARSTATE_FAILURE,
						RRND_STATUS_SIZEOF,
						COMM_TYPE_RRND_STATUS);
				}
				break;
			case INPUT_TYPE_RRND_INFER_CONTACT:
				infer_ct(parser->string);
				break;
			case INPUT_TYPE_RESET_ROUTING_DB:
				LOG("Resetting routing DB");
				comm_send_packet("\x01", 1,
					COMM_TYPE_GENERIC_RESULT);
				/* TODO: Lock! */
				routing_table_free();
				break;
			default:
				parser->basedata->status = PARSER_STATUS_ERROR;
				break;
			}
		} else {
			parser->basedata->status = PARSER_STATUS_ERROR;
		}
		break;
	case INPUT_EXPECT_DATA:
		/* Should not get here... */
		parser->basedata->status = PARSER_STATUS_ERROR;
		break;
	default:
		parser->basedata->status = PARSER_STATUS_ERROR;
		break;
	}
}

void input_parser_reset(struct input_parser *parser)
{
	parser->basedata->status = PARSER_STATUS_GOOD;
	parser->stage = INPUT_EXPECT_DATA_ESCAPE_DELIMITER;
	parser->type = INPUT_TYPE_UNDEFINED;
	free(parser->string);
	parser->string = NULL;
	parser->intdata = 0;
	parser->intdata_index = 0;
}

struct parser *input_parser_init(struct input_parser *parser)
{
	parser->basedata = malloc(sizeof(struct parser));
	if (parser->basedata == NULL)
		return NULL;
	parser->string = NULL;
	input_parser_reset(parser);
	return parser->basedata;
}
