#include <math.h>

#include <rrnd.h>

/* PROBABILITIES */

static float estimate_historical_probability(
	const struct rrnd_contact_history *const history);
static float estimate_satpos_accuracy(const unsigned long pos_age);
static float estimate_gsloc_accuracy(const unsigned char elements);

float rrnd_estimator_estimate_prediction_probability(
	const struct rrnd_contact_history *const history,
	const unsigned long satpos_age_sec,
	const unsigned char gsloc_elements,
	const struct rrnd_availability_pattern *const pattern)
{
	float P_h, P_pos, P_loc, P_avail, P_d, P_pred;

	P_h = estimate_historical_probability(history);

	P_pos = estimate_satpos_accuracy(satpos_age_sec);
	P_loc = estimate_gsloc_accuracy(gsloc_elements);
	if (pattern == NULL)
		P_avail = 1.0f;
	else
		P_avail = rrnd_estimator_estimate_avail_accuracy(pattern);
	P_d = P_pos * P_loc * P_avail;

	P_pred = (
		RRND_ESTIMATOR_Apred * P_h +
		RRND_ESTIMATOR_Bpred * P_d +
		RRND_ESTIMATOR_Cpred * P_h * P_d
	);
	RRND_DBG("P_pred=%f (P_h=%f, P_d=%f, P_pos=%f, P_loc=%f, P_avail=%f)\n",
		P_pred, P_h, P_d, P_pos, P_loc, P_avail);

	return MIN(1.0f, P_pred);
}

static float estimate_historical_probability(
	const struct rrnd_contact_history *const history)
{
	int i;
	float histprob, bitprob;

	if (history->elements == 0 && history->contacts_in_bitmap == 0)
		return RRND_ESTIMATOR_HISTORICAL_INITIAL_PROBABILITY;
	if (history->elements != 0) {
		histprob = 0.0f;
		for (i = 0; i < history->elements; i++)
			histprob += history->prediction_reliability[i];
		histprob /= history->elements;
	} else {
		histprob = 1.0f;
	}
	if (history->contacts_in_bitmap != 0) {
		i = sizeof(history->failed_contacts_bitmap) * 8
			- history->contacts_in_bitmap;
		bitprob = __builtin_popcount(
			(0xFFFFFFFFFFFFFFFFL >> i) &
			history->failed_contacts_bitmap);
		bitprob = 1.0f - (bitprob / history->contacts_in_bitmap);
	} else {
		bitprob = 1.0f;
	}
	return histprob * bitprob;
}

static float estimate_satpos_accuracy(const unsigned long pos_age)
{
#ifdef RRND_ASSUME_SATPOS_CORRECT
	return 1.0f;
#else /* RRND_ASSUME_SATPOS_CORRECT */
	if (pos_age <= RRND_SATPOS_ACCURACY_THRESHOLD)
		return 1.0f;
	return expf(
		-(RRND_SATPOS_ACCURACY_PARAM) *
		((pos_age - RRND_SATPOS_ACCURACY_THRESHOLD) / 1000));
#endif /* RRND_ASSUME_SATPOS_CORRECT */
}

static float estimate_gsloc_accuracy(const unsigned char elements)
{
	if (elements == RRND_LOCATOR_MAX_RECORDS)
		return 1.0f;
	return (RRND_GSLOC_ACCURACY_MINIMUM +
		(1 - RRND_GSLOC_ACCURACY_MINIMUM) /
		(RRND_LOCATOR_MAX_RECORDS - 3) *
		(elements - 3));
}

float rrnd_estimator_estimate_avail_accuracy(
	const struct rrnd_availability_pattern *const pattern)
{
	unsigned char changes =
		rrnd_availability_pattern_get_sig_changes(pattern);
	unsigned char contacts = pattern->contacts_honored;

	if (contacts < 2 || changes < 1 || pattern->period == 0)
		return 1.0f;
	return MAX(RRND_AVAIL_ACCURACY_MINIMUM, 1 - changes / contacts);
}

/* CAPACITY */

/*
 * NOTE
 * For future optimization, see also:
 * http://math.stackexchange.com/questions/102978/
 *     incremental-computation-of-standard-deviation
 * https://en.wikipedia.org/wiki/
 *     Algorithms_for_calculating_variance#Online_algorithm
 */
struct rrnd_capacity_estimation rrnd_estimator_estimate_capacity(
	const unsigned long max_tx_bitrate,
	const struct rrnd_contact_history *const history,
	const unsigned long contact_duration)
{
	int i;
	unsigned long long brsum;
	long long br, sqsum;
	struct rrnd_capacity_estimation ret;

	if (history->elements < RRND_ESTIMATOR_MIN_HISTORY_RECORDS) {
		ret.mean = (unsigned long)((float)max_tx_bitrate *
			RRND_ESTIMATOR_MEAN_BITRATE_FACTOR);
		ret.std_deviation = (float)max_tx_bitrate *
			RRND_ESTIMATOR_DEFAULT_STD_DEVIATION_FACTOR;
	} else {
		brsum = sqsum = 0;
		for (i = 0; i < history->elements; i++)
			brsum += history->tx_bitrate[i];
		ret.mean = brsum / history->elements;
		for (i = 0; i < history->elements; i++) {
			br = (long long)history->tx_bitrate[i];
			sqsum += (br - (long long)ret.mean)
				* (br - (long long)ret.mean);
		}
		ret.std_deviation = sqrtf((float)sqsum / history->elements);
	}
	/*RRND_DBG("Contact BR estimation: %lu; %f\n",*/
	/*	 ret.mean, ret.std_deviation);*/
	ret.mean *= contact_duration / 1000;
	ret.std_deviation *= contact_duration / 1000;
	return ret;
}
