#include <stddef.h>
#include <stdint.h>

#include <upcnFlags.h>

#ifdef INCLUDE_BOARD_LIB

#include <stm32f4xx.h>
#include <misc.h>

#include <hwf4/gpio.h>
#include <hwf4/timer.h>

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <malloc.h>
#include <mini-printf.h>

#include <FreeRTOS.h>
#include <semphr.h>
#include <task.h>

#ifndef USE_STM_SHA1
#include <sha2.h>
#include <hmac_sha2.h>
#endif /* USE_STM_SHA1 */

#include <mpu.h>
#include <comm.h>

#endif /* INCLUDE_BOARD_LIB */

#include <upcn.h>

/*
 * In local builds, most of the functions from this file aren't available.
 */

#ifdef INCLUDE_BOARD_LIB

static SemaphoreHandle_t comm_user_semaphore;

#define DEFAULT_PRINTF_BUFFER_SIZE 192
static char printf_buffer[DEFAULT_PRINTF_BUFFER_SIZE];
static SemaphoreHandle_t printf_semaphore;

static uint8_t upcncomm_init(void)
{
	comm_init();
	comm_user_semaphore = xSemaphoreCreateBinary();
	if (comm_user_semaphore == NULL)
		return 0;
	xSemaphoreGive(comm_user_semaphore);
	printf_semaphore = xSemaphoreCreateBinary();
	if (printf_semaphore == NULL)
		return 0;
	xSemaphoreGive(printf_semaphore);
	return 1;
}

static inline void comm_write_char_escaped(unsigned char current)
{
	char buf[5];

	if (isprint(current)) {
		comm_write_raw(&current, 1);
	} else {
		mini_snprintf(buf, 5, "\\x%02x", current);
		comm_write_raw(buf, 4);
	}
}

void comm_write_raw_escaped(void *data, size_t length)
{
	unsigned char *str;

	str = (unsigned char *)data;
	for (size_t i = 0; i < length; i++)
		comm_write_char_escaped(str[i]);
}

void comm_write_escaped(char *string)
{
	unsigned char *str;

	str = (unsigned char *)string;
	while ((*str) != '\0') {
		comm_write_char_escaped(*str);
		str++;
	}
}

int comm_printf(char *format, ...)
{
	va_list args;
	int length;

	while (pdFALSE == xSemaphoreTake(printf_semaphore, portMAX_DELAY))
		;
	va_start(args, format);
	length = mini_vsnprintf(
		printf_buffer, DEFAULT_PRINTF_BUFFER_SIZE, format, args);
	va_end(args);
	/* If the string to be printed is longer, we should allocate memory */
	/* and do the printing again... [ TODO: removed! ] */
	if (length >= DEFAULT_PRINTF_BUFFER_SIZE - 1) {
		printf_buffer[length - 2] = '~';
		printf_buffer[length - 1] = '\n';
	}
	if (length >= 0)
		comm_write_raw(printf_buffer, length);
	xSemaphoreGive(printf_semaphore);
	return length;
}

/* - LOCK */

void comm_lock(void)
{
	while (pdFALSE == xSemaphoreTake(comm_user_semaphore, portMAX_DELAY))
		;
}

uint8_t comm_try_lock(uint16_t ms_timeout)
{
	return xSemaphoreTake(comm_user_semaphore,
		ms_timeout / portTICK_PERIOD_MS);
}

void comm_unlock(void)
{
	xSemaphoreGive(comm_user_semaphore);
}

/* LED OUTPUT */

/* We define 5 LED presets for each board for debugging purposes. */
/* So we get an abstract interface that works well */
/* on both the STM32F4 discovery and SOMP2 boards. */
/* The 6th preset is for indicating errors. */
#define LED_DBG_PRESET_COUNT 6

#if defined(BOARD_STM32F4_DISCOVERY)

/* The STM32F4 discovery has 4 discretely accessible LEDs. */
#define LED_COUNT 4
#define LED_PINS { pin_d12, pin_d13, pin_d14, pin_d15 }

/* Define the association between debugging presets and LED pins: */
/* 0 .. 1 .. 2 .. 3 .. 4 LEDs on */
/* Each HEX digit represents a mode value for led_pin_set() */
/* Order is <PIN1><PIN2><PIN3><PIN4>, see LED_PINS. */
#define LED_DBG_PRESETS { 0x0000, 0x1000, 0x1100, 0x1110, 0x1111, 0x0010 }

#elif defined(BOARD_SOMP2_v0_2)

/* The SOMP2 v0.2 board has 2 RGB (?) LEDs wired to 3 pins */
/* that can only be controlled in 6 discrete modes. */
#define LED_COUNT 3
#define LED_PINS { pin_c10, pin_c11, pin_c12 }

/* Available LED modes on SOMP2 v0.2: */
/* Each HEX digit represents a mode value for led_pin_set(), */
/* order is <PIN1><PIN2><PIN3>, see LED_PINS. */
enum led_mode {
	LED_MODE_1_RED     = 0x101,
	LED_MODE_2_RED     = 0x011,
	LED_MODE_2_BLUE    = 0x210,
	LED_MODE_BOTH_BLUE = 0x110,
	LED_MODE_2_REDBLUE = 0x010,
	LED_MODE_OFF       = 0x222
};

/* Define the association between debugging presets and LED pins: */
/* off .. one blue .. both blue .. left red .. right red */
#define LED_DBG_PRESETS { LED_MODE_OFF, LED_MODE_2_BLUE, \
	LED_MODE_BOTH_BLUE, LED_MODE_1_RED, LED_MODE_2_RED, LED_MODE_2_RED }

#else

/* Fail if there is no board defined... */
#error "No board defined in upcn.h!"

#endif

struct pin *leds[LED_COUNT]               = LED_PINS;
int led_dbg_presets[LED_DBG_PRESET_COUNT] = LED_DBG_PRESETS;

/* Initializes a GPIO pin */
static void led_pin_init(struct pin *p)
{
	pin_enable(p);
	pin_set_otype(p, PIN_TYPE_PUSHPULL);
	pin_set_ospeed(p, PIN_SPEED_2MHZ);
	pin_set_pupd(p, PIN_PUPD_NONE);
}

/* Sets the specified output mode for the specified GPIO pin. */
/* mode = 0: LOW */
/* mode = 1: HIGH */
/* mode = 2: FLOAT ("analog") */
void led_pin_set(struct pin *p, int mode)
{
	if (mode == 1) { /* set */
		pin_set(p);
		pin_set_mode(p, PIN_MODE_OUTPUT);
	} else if (mode == 0) { /* reset */
		pin_reset(p);
		pin_set_mode(p, PIN_MODE_OUTPUT);
	} else { /* float */
		pin_set_mode(p, PIN_MODE_ANALOG);
	}
}

/* Sets the specified led_mode value by decomposing it */
/* into HEX digits and setting the associated pins. */
/* Example: 0x102 would mean leds[0] = HI, leds[1] = LO, leds[2] = N/C */
static void led_set_mode(int led_mode)
{
	int mask = 0x3;
	int shift = 0;
	int i;

	for (i = LED_COUNT - 1; i != -1; i--) {
		led_pin_set(leds[i], (led_mode & mask) >> shift);
		mask <<= 4;
		shift += 4;
	}
}

/* Initializes the defined LED pins. */
static void led_init(void)
{
	int i;

	for (i = 0; i < LED_COUNT; i++)
		led_pin_init(leds[i]);
}

/* Sets the LED preset, see above. */
void led_set(int led_preset)
{
	if (led_preset < 0 || led_preset >= LED_DBG_PRESET_COUNT)
		return;

	led_set_mode(led_dbg_presets[led_preset]);
}

/* CLOCK */

static uint64_t ref_ticks;
static uint64_t ref_timestamp;

void clock_set_timestamp(const uint64_t timestamp)
{
	ref_ticks = timer_get_ticks();
	ref_timestamp = timestamp;
}

void clock_reset(void)
{
	LOG("Resetting clock!");
	clock_set_timestamp(0);
}

uint64_t clock_get_timestamp(void)
{
	uint64_t tick_difference = timer_get_ticks() - ref_ticks;

	return ref_timestamp + (uint64_t)(tick_difference / 1000000ULL);
}

uint64_t clock_get_timestamp_ms(void)
{
	uint64_t tick_difference = timer_get_ticks() - ref_ticks;

	return ref_timestamp * 1000ULL + (uint64_t)(tick_difference / 1000ULL);
}

uint64_t clock_get_timestamp_us(void)
{
	return ref_timestamp * 1000000ULL + timer_get_ticks() - ref_ticks;
}

uint64_t clock_get_ticks(void)
{
	return timer_get_ticks();
}

/* RNG */

static void random_init(void)
{
	RCC_AHB2PeriphClockCmd(RCC_AHB2Periph_RNG, ENABLE);
	RNG_Cmd(ENABLE);
}

uint32_t random_get(void)
{
	while (RNG_GetFlagStatus(RNG_FLAG_DRDY) == RESET)
		;
	return RNG_GetRandomNumber();
}

/* HMAC */

static void hash_init(void)
{
	RCC_AHB2PeriphClockCmd(RCC_AHB2Periph_HASH, ENABLE);
}

void hash(
	uint8_t *message, uint32_t message_length,
	uint8_t hash[UPCN_HASH_LENGTH])
{
#ifdef USE_STM_SHA1
	ASSERT(HASH_SHA1(message, message_length, hash));
#else /* USE_STM_SHA1 */
	sha256(message, message_length, hash);
#endif /* USE_STM_SHA1 */
}

void hash_hmac(
	uint8_t *key, uint32_t key_length,
	uint8_t *message, uint32_t message_length,
	uint8_t hash[UPCN_HASH_LENGTH])
{
#ifdef USE_STM_SHA1
	ASSERT(HMAC_SHA1(key, key_length, message, message_length, hash));
#else /* USE_STM_SHA1 */
	hmac_sha256(key, key_length, message, message_length,
		hash, UPCN_HASH_LENGTH);
#endif /* USE_STM_SHA1 */
}

/* CRC */

static void crc_init(void)
{
	RCC_AHB1PeriphClockCmd(RCC_AHB1Periph_CRC, ENABLE);
}

uint32_t crc32(uint32_t *input, size_t len)
{
	CRC->CR = CRC_CR_RESET;
	while (len--)
		CRC->DR = *(input++);
	return CRC->DR;
}

uint32_t crc32_8(uint8_t *input, size_t len)
{
	CRC->CR = CRC_CR_RESET;
	while (len--)
		CRC->DR = *(input++);
	return CRC->DR;
}

/* INIT */

void firmware_init(void)
{
	/* This is done by the reset handler run before main() */
	/* SystemInit(); */
	/* XXX: FreeRTOS documentation says this is important... */
	/* Ensure all priority bits are assigned as preemption priority bits. */
	NVIC_PriorityGroupConfig(NVIC_PriorityGroup_4);
	timer_init();
	clock_set_timestamp(0);
	upcncomm_init();
	led_init();
	/* Enable additional fault handlers */
	SCB->SHCSR |= SCB_SHCSR_USGFAULTENA_Msk | SCB_SHCSR_BUSFAULTENA_Msk;
	mpu_init();
	random_init();
	hash_init();
	crc_init();
}

/* DBG INFO */
void upcn_print_info(void)
{
	static char wbuf[16];
	struct mallinfo mi = mallinfo();

	comm_lock();
	comm_write("\nSTATE INFO\n==========\n\n");
	comm_write("Current time:        ");
	sprintu64(wbuf, clock_get_timestamp());
	comm_write(wbuf);
	comm_write("\nTotal mem allocated: ");
	sprintu32(wbuf, mi.uordblks);
	comm_write(wbuf);
	comm_write(" bytes\nTotal mem free:      ");
	sprintu32(wbuf, mi.fordblks);
	comm_write(wbuf);
	comm_write(" bytes\nTotal mem in pool:   ");
	sprintu32(wbuf, mi.arena);
	comm_write(wbuf);
	comm_write(" bytes\n");
	comm_unlock();
}

/* RESET */
void upcn_reset_system(void)
{
	uint64_t btime;

	vTaskSuspendAll();
	while (comm_output_waiting())
		;
	btime = clock_get_timestamp_ms();
	while (clock_get_timestamp_ms() - btime <= 200)
		;
	NVIC_SystemReset();
}

/* CPU Cycle debugging, see Cortex-M3 ref. manual, p. 11-15 ff. */

#define DWT_CTRL 0xE0001000UL
#define DWT_CYCCNT 0xE0001004UL
#define NUMCOMP 0x40000000UL
#define CYCCNTENA 0x1UL

#define DWT_CycCntEna()   ((*((uint32_t *)0xE0001000)) = (NUMCOMP | CYCCNTENA))
#define DWT_CycCntDis()   ((*((uint32_t *)0xE0001000)) = NUMCOMP)
#define DWT_CycCntRes()   ((*((uint32_t *)0xE0001004)) = 0)
#define DWT_CycCntGet()   (*((uint32_t *)0xE0001004))

void cyccnt_start(void)
{
#ifdef DEBUG
	DWT_CycCntRes();
	DWT_CycCntEna();
#endif /* DEBUG */
}

uint32_t cyccnt_stop(void)
{
	uint32_t result;

#ifdef DEBUG
	DWT_CycCntDis();
	result = DWT_CycCntGet();
#else /* DEBUG */
	result = 0;
#endif /* DEBUG */
	return result;
}

#endif /* INCLUDE_BOARD_LIB */

/* COMMON FUNCTIONS */

static const char digits[] = "0123456789abcdef";

#define sprintu_generic(T, cur, base, num) do { \
	T tmp = num; \
	do { \
		++cur; \
		tmp /= base; \
	} while (tmp); \
	*cur = '\0'; \
	do { \
		--cur; \
		*cur = digits[num % base]; \
		num /= base; \
	} while (num); \
} while (0)

char *sprintu32(char *out, uint32_t num)
{
	char *cur = out;

	sprintu_generic(uint32_t, cur, 10LL, num);
	return out;
}

char *sprintu32x(char *out, uint32_t num)
{
	char *cur = out;

	*(cur++) = '0';
	*(cur++) = 'x';
	sprintu_generic(uint32_t, cur, 16LL, num);
	return out;
}

char *sprintu64(char *out, uint64_t num)
{
	char *cur = out;

	sprintu_generic(uint64_t, cur, 10LL, num);
	return out;
}

char *sprintu64x(char *out, uint64_t num)
{
	char *cur = out;

	*(cur++) = '0';
	*(cur++) = 'x';
	sprintu_generic(uint64_t, cur, 16LL, num);
	return out;
}
