#ifndef UPCN_H_INCLUDED
#define UPCN_H_INCLUDED

#include <stddef.h>
#include <stdint.h>
#include <config.h>
#include <upcnFlags.h>

enum task_tags {
	IDLE_TASK_TAG = 0,
	INPUT_TASK_TAG,
	ROUTER_TASK_TAG,
	BUNDLE_PROCESSOR_TASK_TAG,
	CONTACT_MANAGER_TASK_TAG,
	GROUND_STATION_TASK_TAG,
	ROUTER_OPTIMIZER_TASK_TAG,
	FS_TASK_TAG
};


enum upcn_result {
	UPCN_OK,
	UPCN_FAIL
};

enum convergence_layer {
	CLA_UNKNOWN,
	CLA_UDP,
	CLA_AX25
	/* ... */
};

/* COMMON FUNCTIONS */

#define count_list_next_length(list, cnt) do { \
	if (list != NULL) \
		cnt = 1; \
	iterate_list_next(list) { ++cnt; } \
} while (0)

#define iterate_list_next(list) \
	for (; list; list = (list)->next)

#define list_element_free(list) do { \
	void *e = list->next; \
	free(list); \
	list = e; \
} while (0)

#define list_free(list) { \
while (list != NULL) { \
	list_element_free(list); \
} }

#ifdef USE_STM_SHA1
#define UPCN_HASH_LENGTH 20
#else /* USE_STM_SHA1 */
#define UPCN_HASH_LENGTH 32
#endif /* USE_STM_SHA1 */

#ifdef INCLUDE_BOARD_LIB

void firmware_init(void);

void clock_reset(void);
void clock_set_timestamp(const uint64_t timestamp);

uint64_t clock_get_timestamp(void);
uint64_t clock_get_timestamp_ms(void);
uint64_t clock_get_timestamp_us(void);
uint64_t clock_get_ticks(void);

#define clock_set_unix_timestamp(timestamp) \
	clock_set_timestamp(timestamp - DTN_TIMESTAMP_OFFSET)
#define clock_get_unix_timestamp() \
	(clock_get_timestamp() + DTN_TIMESTAMP_OFFSET)
#define clock_get_unix_timestamp_ms() \
	(clock_get_timestamp_ms() + (DTN_TIMESTAMP_OFFSET * 1000))

#include <comm.h>

void comm_lock(void);
uint8_t comm_try_lock(uint16_t ms_timeout);
void comm_unlock(void);

void comm_write_escaped(char *string);
void comm_write_raw_escaped(void *data, size_t length);
int comm_printf(char *format, ...);

void led_set(int led_mode);

/* Helper functions / macros */

char *sprintu32(char *out, uint32_t num);
char *sprintu32x(char *out, uint32_t num);
char *sprintu64(char *out, uint64_t num);
char *sprintu64x(char *out, uint64_t num);

uint32_t random_get(void);

void hash(
	uint8_t *message, uint32_t message_length,
	uint8_t hash[UPCN_HASH_LENGTH]);
void hash_hmac(
	uint8_t *key, uint32_t key_length,
	uint8_t *message, uint32_t message_length,
	uint8_t hash[UPCN_HASH_LENGTH]);

uint32_t crc32(uint32_t *input, size_t len);
uint32_t crc32_8(uint8_t *input, size_t len);

void upcn_print_info(void);

void upcn_reset_system(void);

void cyccnt_start(void);
uint32_t cyccnt_stop(void);

#else /* INCLUDE_BOARD_LIB */

#define firmware_init() ((void)0)

#endif /* INCLUDE_BOARD_LIB */

#ifdef MIN
#undef MIN
#endif

#ifdef MAX
#undef MAX
#endif

#define MIN(a, b) ({ \
	__typeof__(a) _a = (a); \
	__typeof__(b) _b = (b); \
	_b < _a ? _b : _a; \
})
#define MAX(a, b) ({ \
	__typeof__(a) _a = (a); \
	__typeof__(b) _b = (b); \
	_b > _a ? _b : _a; \
})

#define HAS_FLAG(value, flag) ((value & flag) != 0)

/* ASSERT */

#ifdef INCLUDE_ASSERT

#ifndef UPCN_LOCAL

#define ASSERT(value) do { \
	if (!(value)) { \
		led_set(5); \
		for (int i = 0; i < 1000000; i++) \
			; \
		upcn_reset_system(); \
		for (;;) \
			; \
	} \
} while (0)

#else /* UPCN_LOCAL */

#include <stdio.h>
#include <signal.h>

#define ASSERT(value) do { \
	if (!(value)) { \
		printf("Assertion failed in %s on line %d\n", \
			__FILE__, __LINE__); \
		raise(SIGINT); \
		for (;;) \
			; \
	} \
} while (0)

#endif /* UPCN_LOCAL */

#else /* INCLUDE_ASSERT */

#define ASSERT(value) ((void)(value))

#endif /* INCLUDE_ASSERT */

/* Memory debugging */

#ifdef MEMDEBUG

#define malloc(size) upcn_dbg_malloc(size, __FILE__, __LINE__)
#define calloc(num, size) upcn_dbg_calloc(num, size, __FILE__, __LINE__)
#define realloc(ptr, size) upcn_dbg_realloc(ptr, size, __FILE__, __LINE__)
#define free(ptr) do { upcn_dbg_free(ptr, __FILE__, __LINE__); \
	ptr = NULL; } while (0)

void *upcn_dbg_malloc(size_t size, char *file, int line);
void *upcn_dbg_calloc(size_t num, size_t size, char *file, int line);
void *upcn_dbg_realloc(void *ptr, size_t size, char *file, int line);
void upcn_dbg_free(void *ptr, char *file, int line);
void upcn_dbg_memprint(void);

void upcn_dbg_mem_lock(void *ptr);
void upcn_dbg_mem_unlock(void *ptr);

uint32_t upcn_dbg_memstat_get_cur(void);
uint32_t upcn_dbg_memstat_get_max(void);
void upcn_dbg_memstat_reset(void);
void upcn_dbg_memstat_print(void);

#else /* MEMDEBUG */

#define upcn_dbg_memprint(a) ((void)0)

#define upcn_dbg_mem_lock(a) ((void)0)
#define upcn_dbg_mem_unlock(a) ((void)0)

#define upcn_dbg_memstat_get_cur() (0)
#define upcn_dbg_memstat_get_max() (0)
#define upcn_dbg_memstat_reset() ((void)0)
#define upcn_dbg_memstat_print() ((void)0)

#endif /* MEMDEBUG */

/* Logging */

#define LOG_NO_ITEM 0xFFFFFFFF
#define LOGI(message, itemid) LOGA(message, 0xFF, itemid)
#define LOG(message) LOGI(message, LOG_NO_ITEM)

#ifdef LOGGING

#define LOGA(message, actionid, itemid) \
	upcn_dbg_log(message, actionid, (uint32_t)itemid, __FILE__, __LINE__)

void upcn_dbg_log(char *msg, uint8_t actid, uint32_t itemid,
	char *file, int line);
void upcn_dbg_printlogs(void);
void upcn_dbg_clearlogs(void);
void upcn_dbg_log_disable_output(void);
void upcn_dbg_log_enable_output(void);

#else /* LOGGING */

#ifdef LOG_PRINT

extern uint8_t _upcn_dbg_log_disabled;

#define LOGA(message, actionid, itemid) do { \
if (_upcn_dbg_log_disabled) \
	break; \
uint32_t ii = (uint32_t)(itemid); \
comm_printf( \
	"%d: %s (%d) [%s:%d]\n", \
	(int)(clock_get_timestamp_ms()), \
	message, \
	((uint8_t)(actionid) == 0xFF) \
		? (ii == LOG_NO_ITEM ? -1 : (int)itemid) \
		: (int)(actionid), \
	__FILE__, \
	(int)(__LINE__) \
); \
} while (0)

void upcn_dbg_log_disable_output(void);
void upcn_dbg_log_enable_output(void);

#else /* LOG_PRINT */

#define LOGA(message, actionid, itemid) ((void)(itemid))

#define upcn_dbg_log_disable_output() ((void)0)
#define upcn_dbg_log_enable_output() ((void)0)

#endif /* LOG_PRINT */

#define upcn_dbg_printlogs() ((void)0)
#define upcn_dbg_clearlogs() ((void)0)

#endif /* LOGGING */

#ifdef DEBUG_FREERTOS_TRACE

void upcn_dbg_printtrace(void);
void upcn_dbg_resettrace(void);
void upcn_dbg_storetrace(void);
void upcn_dbg_cleartraces(void);
void upcn_dbg_printtraces(void);

#else /* DEBUG_FREERTOS_TRACE */

#define upcn_dbg_printtrace() ((void)0)
#define upcn_dbg_resettrace() ((void)0)
#define upcn_dbg_storetrace() ((void)0)
#define upcn_dbg_cleartraces() ((void)0)
#define upcn_dbg_printtraces() ((void)0)

#endif /* DEBUG_FREERTOS_TRACE */

#ifdef UPCN_TEST_BUILD
void upcntest_print(void);
#endif /* UPCN_TEST_BUILD */

#endif /* UPCN_H_INCLUDED */
