#include <string.h>
#include <limits.h>
#include <math.h>

#include <upcn.h>
#include <rrnd.h>
#include <vec3d.h>
#include <ellipsoidal_coords.h>

#define T_CORR(station) ( \
	RRND_CONTACT_TIME_CORRECTION(((station)->beacon_period == 0) \
		? RRND_DEFAULT_BEACON_PERIOD : (station)->beacon_period * 10))

static void rrnd_try_locate(struct rrnd_gs_info *const station,
	struct vec3d (*const get_satpos)(const unsigned long long time));

static enum rrnd_status cleanup_predictions(
	struct rrnd_predictions *const pred,
	struct rrnd_contact_history *const history,
	const unsigned long long time);

static enum rrnd_status append_prediction(
	struct rrnd_predictions *const pred,
	const struct rrnd_contact_info p);

static enum rrnd_status rrnd_update_contact_history_and_predictions(
	struct rrnd_gs_info *const gs,
	struct vec3d (*const get_satpos)(const unsigned long long time));

static int out_of_last_contact(
	struct rrnd_gs_info *const station, const unsigned long long time);

static void update_station_from_beacon(
	struct rrnd_gs_info *const station,
	const struct rrnd_beacon *const beacon);

static double get_earth_radius(const struct rrnd_gs_info *const gs);

static void append_to_history(struct rrnd_contact_history *const h,
	const float match, const unsigned long br);

static double estimate_start_elevation(const struct rrnd_gs_info *const gs);

static void update_bitrate_estimation(
	struct rrnd_gs_info *const station,
	const unsigned long our_estimation, const unsigned long gs_estimation);

enum rrnd_status rrnd_check_and_finalize(
	struct rrnd_gs_info *const station, const unsigned long long cur_time,
	struct vec3d (*const get_satpos)(const unsigned long long time))
{
	enum rrnd_status rc = RRND_STATUS_UNCHANGED;

	if (out_of_last_contact(station, cur_time) &&
		HAS_FLAG(station->flags, RRND_GS_FLAG_IN_CONTACT)
	) {
		rrnd_locator_add(&station->locator_data,
			station->last_contact_end_time + T_CORR(station),
			get_satpos);
		station->flags &= ~RRND_GS_FLAG_IN_CONTACT;
		rc |= RRND_STATUS_UPDATED | RRND_UPDATE_LOCATOR_DATA;
		rrnd_try_locate(station, get_satpos);
		rc |= rrnd_update_contact_history_and_predictions(
			station, get_satpos);
		rc |= cleanup_predictions(
			&station->predictions, &station->contact_history,
			cur_time);
	}
	return rc;
}

enum rrnd_status rrnd_process(
	struct rrnd_gs_info *const station, const struct rrnd_beacon beacon,
	const unsigned long long time, const unsigned long br_estimation,
	struct vec3d (*const get_satpos)(const unsigned long long time))
{
	enum rrnd_status rc = RRND_STATUS_SUCCESS;

	rc |= rrnd_check_and_finalize(station, time, get_satpos);
	if (HAS_FLAG(rc, RRND_STATUS_FAILED))
		return rc;
	if (!HAS_FLAG(station->flags, RRND_GS_FLAG_IN_CONTACT)) {
		/* Start contact */
		station->flags |= RRND_GS_FLAG_IN_CONTACT;
		if (HAS_FLAG(station->flags, RRND_GS_FLAG_LOCATION_VALID) &&
				rrnd_validator_validate_contact_time(
					time, station->location,
					&station->predictions,
					get_satpos).success == 0)
			return rc | RRND_STATUS_FAILED | RRND_FAILURE_VALIDATOR;
		update_station_from_beacon(station, &beacon);
		if (br_estimation != 0 && br_estimation != ULONG_MAX)
			station->contact_mean_bitrate = br_estimation;
		rrnd_locator_add(&station->locator_data,
			time - T_CORR(station), get_satpos);
		rc |= RRND_UPDATE_LOCATOR_DATA;
		rrnd_try_locate(station, get_satpos);
		if (HAS_FLAG(station->flags,
				RRND_GS_FLAG_INTERMITTENT_AVAILABILITY)) {
			rrnd_availability_pattern_update(
				&station->availability_pattern, time,
				beacon.availability_duration,
				station->last_contact_end_time, 1);
			rc |= RRND_UPDATE_AVAILABILITY;
		}
		station->beacon_count = 0;
		station->last_contact_start_time = time;
	} else if (HAS_FLAG(station->flags,
			RRND_GS_FLAG_INTERMITTENT_AVAILABILITY)) {
		rrnd_availability_pattern_update(
			&station->availability_pattern, time,
			beacon.availability_duration,
			station->last_contact_end_time, 0);
		rc |= RRND_UPDATE_AVAILABILITY;
	}
	update_bitrate_estimation(station, br_estimation, beacon.rx_bitrate);
	station->beacon_count++;
	station->last_contact_end_time = time;
	rc |= RRND_STATUS_UPDATED;
	return rc;
}

enum rrnd_status rrnd_infer_contact(
	struct rrnd_gs_info *const station, const unsigned long long cur_time,
	struct vec3d (*const get_satpos)(const unsigned long long time),
	const unsigned long satpos_age_sec)
{
	enum rrnd_status rc = RRND_STATUS_SUCCESS;
	unsigned long long start_time;
	struct rrnd_contact_info result;
	struct rrnd_availability_pattern *ap = NULL;

	rc |= rrnd_check_and_finalize(station, cur_time, get_satpos);
	if (HAS_FLAG(rc, RRND_STATUS_FAILED))
		return rc;
	rrnd_try_locate(station, get_satpos);
	if (!HAS_FLAG(station->flags, RRND_GS_FLAG_LOCATION_VALID))
		return rc | RRND_STATUS_FAILED | RRND_FAILURE_LOCATOR;
	if (station->predictions.elements != 0)
		start_time = (
			station->predictions.contacts[
				station->predictions.elements - 1].start +
			station->predictions.contacts[
				station->predictions.elements - 1].duration +
			1);
	else
		start_time = station->last_contact_end_time;
	start_time = MAX(start_time, cur_time);
	if (HAS_FLAG(station->flags, RRND_GS_FLAG_INTERMITTENT_AVAILABILITY))
		ap = &station->availability_pattern;
	if (rrnd_predictor_predict(
			&result, station->location, start_time,
			estimate_start_elevation(station),
			RRND_CONTACT_MIN_DURATION, RRND_PREDICTION_TIME_STEP,
			ap, get_satpos) != RRND_SUCCESS)
		return rc | RRND_STATUS_FAILED | RRND_FAILURE_PREDICTOR;
	result.capacity = rrnd_estimator_estimate_capacity(
		station->tx_bitrate, &station->contact_history,
		result.duration);
	result.probability = rrnd_estimator_estimate_prediction_probability(
		&station->contact_history, satpos_age_sec,
		station->locator_data.elements, ap);
	rc |= append_prediction(&station->predictions, result);
	return rc;
}

static void rrnd_try_locate(struct rrnd_gs_info *const station,
	struct vec3d (*const get_satpos)(const unsigned long long time))
{
	if (rrnd_locator_locate(&station->location,
		HAS_FLAG(station->flags, RRND_GS_FLAG_LOCATION_VALID),
		&station->locator_data, get_earth_radius(station),
		RRND_START_ELEVATION, get_satpos)
			== RRND_SUCCESS
	) {
		station->flags |= RRND_GS_FLAG_LOCATION_VALID;
	} else {
		station->flags &= ~RRND_GS_FLAG_LOCATION_VALID;
	}
}

static enum rrnd_status cleanup_predictions(
	struct rrnd_predictions *const pred,
	struct rrnd_contact_history *const history,
	const unsigned long long time)
{
	int c;

	for (c = 0; c < pred->elements; c++) {
		if ((pred->contacts[c].start + pred->contacts[c].duration)
				> time)
			break;
	}
	/* NOTE: Everything that was still in here is a FAILED prediction */
	if (c != 0) {
		history->failed_contacts_bitmap <<= c;
		history->failed_contacts_bitmap |=
			(unsigned long long)(0xFFFFFFFF >> (32 - c));
		history->contacts_in_bitmap += c;
		if (history->contacts_in_bitmap > RRND_CONTACT_SUCCESS_HISTORY)
			history->contacts_in_bitmap =
				RRND_CONTACT_SUCCESS_HISTORY;
		pred->elements -= c;
		if (pred->elements != 0)
			memmove(&pred->contacts[0], &pred->contacts[c],
				pred->elements *
				sizeof(struct rrnd_contact_info));
		return RRND_STATUS_UPDATED | RRND_UPDATE_CONTACTS;
	}
	return RRND_STATUS_UNCHANGED;
}

static enum rrnd_status append_prediction(
	struct rrnd_predictions *const pred,
	const struct rrnd_contact_info p)
{
	pred->last = p;
	if (pred->elements >= RRND_CONTACT_PREDICTION_LENGTH)
		return RRND_STATUS_UPDATED | RRND_FAILURE_OVERFLOW;
	pred->contacts[pred->elements] = p;
	pred->elements++;
	return RRND_STATUS_UPDATED | RRND_UPDATE_CONTACTS;
}

static enum rrnd_status rrnd_update_contact_history_and_predictions(
	struct rrnd_gs_info *const gs,
	struct vec3d (*const get_satpos)(const unsigned long long time))
{
	enum rrnd_status rc = RRND_STATUS_UNCHANGED;
	struct rrnd_validation_result vr;
	struct rrnd_contact_history *h;
	struct rrnd_predictions *p;

	if (!HAS_FLAG(gs->flags, RRND_GS_FLAG_LOCATION_VALID))
		return RRND_STATUS_UNCHANGED;
	vr = rrnd_validator_validate_contact(
		gs->last_contact_start_time, gs->last_contact_end_time,
		gs->location, &gs->predictions, get_satpos);
	h = &gs->contact_history;
	h->failed_contacts_bitmap <<= 1;
	h->failed_contacts_bitmap |= !vr.success;
	if (++(h->contacts_in_bitmap) > RRND_CONTACT_SUCCESS_HISTORY)
		h->contacts_in_bitmap = RRND_CONTACT_SUCCESS_HISTORY;
	rc |= RRND_STATUS_UPDATED;
	/* NOTE We do not fail here as it was the _last_ contact which failed */
	/* if (!vr.success) */
	/* return rc | RRND_STATUS_FAILED | RRND_FAILURE_VALIDATOR; */
	if (vr.prediction != -1) {
		/* Append match to history */
		append_to_history(h, vr.match, gs->contact_mean_bitrate);
		/* Remove prediction from list */
		p = &gs->predictions;
		if (vr.prediction < p->elements - 1)
			memmove(&p->contacts[vr.prediction],
				&p->contacts[vr.prediction + 1],
				(p->elements - vr.prediction) *
					sizeof(struct rrnd_contact_info));
		p->elements--;
		rc |= RRND_STATUS_UPDATED | RRND_UPDATE_CONTACTS;
	}
	return rc;
}

/* Helpers */

static int out_of_last_contact(
	struct rrnd_gs_info *const station, const unsigned long long time)
{
	return (
		time > station->last_contact_end_time &&
		(time - station->last_contact_end_time)
			> RRND_BEACON_MAX_WAIT
	);
}

static void update_station_from_beacon(
	struct rrnd_gs_info *const station,
	const struct rrnd_beacon *const beacon)
{
	station->flags |= (beacon->flags & 0x0F);
	station->tx_bitrate = beacon->tx_bitrate;
	station->rx_bitrate = beacon->rx_bitrate;
	station->contact_mean_bitrate = beacon->rx_bitrate;
	station->beacon_period = beacon->period;
}

static double get_earth_radius(const struct rrnd_gs_info *const gs)
{
	double Rearth = RRND_EARTH_RADIUS;

	if (HAS_FLAG(gs->flags, RRND_GS_FLAG_LOCATION_VALID)) {
		Rearth = coord_earth_radius_at(gs->location);
		/* max diff. from mean radius is around 15 km */
		if (ABS(Rearth - RRND_EARTH_RADIUS) > 16)
			Rearth = RRND_EARTH_RADIUS;
	}
	return Rearth;
}

static void append_to_history(struct rrnd_contact_history *const h,
	const float match, const unsigned long br)
{
	if (h->elements < RRND_CONTACT_HISTORY_LENGTH)
		h->elements++;
	h->prediction_reliability[h->index] = match;
	h->tx_bitrate[h->index] = br;
	h->index++;
	if (h->index == RRND_CONTACT_HISTORY_LENGTH)
		h->index = 0;
}

static double estimate_start_elevation(const struct rrnd_gs_info *const gs)
{
	double elev_mean;

#ifdef RRND_PREDICTION_ESTIMATE_START_ELEVATION
	if (!HAS_FLAG(gs->flags, RRND_GS_FLAG_LOCATION_VALID) ||
			gs->last_contact_start_time == 0 ||
			gs->last_contact_end_time == 0)
		return RRND_START_ELEVATION;
	elev_mean = rrnd_locator_get_mean_start_elevation(
		&gs->locator_data, gs->location,
		RRND_PREDICTION_ESTIMATE_START_ELEVATION);
	if (isnan(elev_mean) ||
			elev_mean < RRND_VALIDATOR_MIN_VALIDATION_ELEVATION)
		return RRND_START_ELEVATION;
	return elev_mean;
#else /* RRND_PREDICTION_ESTIMATE_START_ELEVATION */
	return RRND_START_ELEVATION;
#endif /* RRND_PREDICTION_ESTIMATE_START_ELEVATION */
}

static void update_bitrate_estimation(
	struct rrnd_gs_info *const station,
	const unsigned long our_estimation, const unsigned long gs_estimation)
{
	unsigned long est = our_estimation, br;
	/* Should never get 0 */
	unsigned short n_beacon = station->beacon_count + 1;

	if (our_estimation == 0 || our_estimation == ULONG_MAX)
		est = gs_estimation;
	/* Online Mean Value */
	br = station->contact_mean_bitrate;
	br = (est + (n_beacon - 1) * br) / n_beacon;
	station->contact_mean_bitrate = br;
	/* Max RX estimation from GS */
	station->rx_bitrate = MAX(station->rx_bitrate, gs_estimation);
}

float rrnd_interval_intersect(
	const unsigned long long orig_start, const unsigned long long orig_end,
	const unsigned long long new_start, const unsigned long long new_end)
{
	unsigned long long diff;

	if (
		orig_start >= new_end ||
		new_start >= orig_end ||
		orig_end <= orig_start ||
		new_end <= new_start
	) {
		return 0.0f;
	}
	if (orig_start <= new_start) {
		if (new_end >= orig_end)
			diff = orig_end - new_start;
		else
			diff = new_end - new_start;
	} else {
		if (orig_end >= new_end)
			diff = new_end - orig_start;
		else
			diff = orig_end - orig_start;
	}
	return (float)diff / (float)(orig_end - orig_start);
}
