#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <upcn/upcn.h>
#include <upcn/satpos.h>
#include <upcn/rrnd.h>
#include <upcn/json_escape.h>

#include "common.h"

#define MSTIME(t) ((unsigned long long)(t) * 1000ULL)

/* As everything is static, it is initialized by zero */
static char gslist[ND_LOCAL_MAX_GS][ND_LOCAL_MAX_EID_LENGTH];
static struct rrnd_gs_info gsinfo[ND_LOCAL_MAX_GS];
static unsigned long gscount;

static int routing_db_local_get(
	struct rrnd_gs_info *const r, const char *const eid);
static int routing_db_local_update_gs(const struct rrnd_gs_info info);
static int routing_db_local_delete_gs(const struct rrnd_gs_info info);

void nd_cmd_process_locally(
	const struct nd_command *const cmd, struct nd_result *const res)
{
	struct rrnd_gs_info gs;

	res->rc = RRND_STATUS_FAILED;
	res->gs_data_json[0] = '\0';
	switch (cmd->type) {
	case ND_COMMAND_TEST_CONNECTIVITY:
		res->rc = RRND_STATUS_SUCCESS;
		break;
	case ND_COMMAND_INIT:
		if (cmd->string_data[0] == '\0' || cmd->int_data == LLONG_MAX)
			break;
		satpos_init(cmd->string_data, MSTIME(cmd->int_data));
		if (satpos_last_result() == SATPOS_SUCCESS) {
			res->rc = RRND_STATUS_SUCCESS;
			puts("> SATPOS initialized successfully");
		}
		break;
	case ND_COMMAND_RESET:
		gscount = 0;
		res->rc = RRND_STATUS_SUCCESS;
		puts("> RRND GS records reset");
		break;
	case ND_COMMAND_PROCESS_BEACON:
		if (cmd->string_data[0] == '\0' || cmd->int_data == LLONG_MAX)
			break;
		if (routing_db_local_get(&gs, cmd->string_data) != EXIT_SUCCESS)
			break;
		res->rc = rrnd_process(
			&gs, cmd->beacon, MSTIME(cmd->int_data), 0, satpos_get);
		if (HAS_FLAG(res->rc, RRND_STATUS_UPDATED))
			routing_db_local_update_gs(gs);
		else if (HAS_FLAG(res->rc, RRND_STATUS_DELETE))
			routing_db_local_delete_gs(gs);
		break;
	case ND_COMMAND_INFER_CONTACT:
		if (cmd->string_data[0] == '\0' || cmd->int_data == LLONG_MAX)
			break;
		if (routing_db_local_get(&gs, cmd->string_data) != EXIT_SUCCESS)
			break;
		res->rc = rrnd_infer_contact(
			&gs, MSTIME(cmd->int_data), satpos_get,
			satpos_get_age(MSTIME(cmd->int_data)));
		if (HAS_FLAG(res->rc, RRND_STATUS_UPDATED))
			routing_db_local_update_gs(gs);
		break;
	case ND_COMMAND_GET_GS:
		if (cmd->string_data[0] == '\0')
			break;
		if (routing_db_local_get(&gs, cmd->string_data) != EXIT_SUCCESS)
			break;
		printf("> RRND GS record queried: %s\n",
			cmd->string_data);
		rrnd_print_gs_info(res->gs_data_json, ND_GS_JSON_STR_SIZE,
			cmd->string_data, &gs);
		res->rc = RRND_STATUS_SUCCESS;
		break;
	case ND_COMMAND_RESET_PERF:
		res->rc = RRND_STATUS_SUCCESS;
		break;
	case ND_COMMAND_STORE_PERF:
		printf("> Memory used: %lu\n",
			sizeof(struct rrnd_gs_info) * gscount);
		res->rc = RRND_STATUS_SUCCESS;
		break;
	default:
		break;
	}
}

static int routing_db_local_get(
	struct rrnd_gs_info *const r, const char *const eid)
{
	unsigned long i;

	for (i = 0; i < gscount; i++) {
		if (strcmp(gslist[i], eid) == 0) {
			*r = gsinfo[i];
			return EXIT_SUCCESS;
		}
	}
	if (gscount >= ND_LOCAL_MAX_GS)
		return EXIT_FAILURE;
	if (strlen(eid) >= ND_LOCAL_MAX_EID_LENGTH)
		return EXIT_FAILURE;
	strncpy(gslist[gscount], eid, ND_LOCAL_MAX_EID_LENGTH);
	memset(&gsinfo[gscount], 0, sizeof(struct rrnd_gs_info));
	gsinfo[gscount].gs_reference = (void *)gscount;
	*r = gsinfo[gscount++];
	return EXIT_SUCCESS;
}

static int routing_db_local_update_gs(const struct rrnd_gs_info info)
{
	unsigned long ref = (unsigned long)info.gs_reference;

	if (ref >= gscount)
		return EXIT_FAILURE;
	gsinfo[ref] = info;
	return EXIT_SUCCESS;
}

static int routing_db_local_delete_gs(const struct rrnd_gs_info info)
{
	unsigned long ref = (unsigned long)info.gs_reference;

	if (ref >= gscount)
		return EXIT_FAILURE;
	/* TODO */
	return EXIT_SUCCESS;
}
