import os
import json
import argparse
import logging
from .nd.client import NDClient
from .s1_static_scenario import StaticDiscoveryScenario
from .s2_intermittent_scenario import IntermittentAvailabilityScenario
from .s3_probability_test import ProbabilityInferenceScenario
from .s4_altitude_test import CriticalAltitudeTestScenario
from .s5_combined_test import CombinedTestScenario
from .s6_scalability_test import ScalabilityTestScenario
from .s7_short_test import ShortTestScenario
from .s8_capacity_test import CapacityTestScenario

SCENARIOS = {
    "1": StaticDiscoveryScenario,
    "2": IntermittentAvailabilityScenario,
    "3": ProbabilityInferenceScenario,
    "4": CriticalAltitudeTestScenario,
    "5": CombinedTestScenario,
    "6": ScalabilityTestScenario,
    "7": ShortTestScenario,
    "8": CapacityTestScenario
}

LIBRARIES = ["pyephem", "skyfield"]
DEFAULT_LIBRARY = "pyephem"

env_library = os.environ.get("RRND_LIBRARY", DEFAULT_LIBRARY)
if env_library not in LIBRARIES:
    env_library = DEFAULT_LIBRARY

VERBOSE_LEVELS = {"0": 0, "1": 1, "2": 2, "3": 2, "error": 0, "warn": 0,
                  "warning": 0, "info": 1, "debug": 2,
                  "v": 1, "vv": 2, "vvv": 2, "": 1}
DEFAULT_VERBOSE_LEVEL = 1

env_verbose_level = VERBOSE_LEVELS.get(
    os.environ.get("RRND_LOGLEVEL", str(DEFAULT_VERBOSE_LEVEL))
    .strip().lower(), DEFAULT_VERBOSE_LEVEL)

env_runs = os.environ.get("RRND_RUNS", "0")
if not env_runs.isdigit():
    env_runs = 0
else:
    env_runs = int(env_runs)

parser = argparse.ArgumentParser(description="Run a test scenario.")
parser.add_argument(
    "-t", "--scenario", choices=sorted(SCENARIOS.keys()), required=True,
    help="the selected test scenario")
parser.add_argument(
    "-l", "--library", choices=LIBRARIES, default=env_library,
    help="the astronomic library to be used")
parser.add_argument(
    "-z", "--zmq",
    default=os.environ.get("RRND_PROXY", "tcp://127.0.0.1:7763"),
    help="the ZMQ address (of rrnd_proxy) to connect to")
parser.add_argument(
    "-f", "--file", type=argparse.FileType("w"), default=None,
    help="the file for storing collected data")
parser.add_argument(
    "-r", "--runs", type=int, default=env_runs,
    help="the run count, defaults to a scenario-specific value")
parser.add_argument(
    "-p", "--param", type=int, default=None,
    help="a scenario parameter, if the selected scenario supports it")
parser.add_argument(
    "-v", "--verbose", action="count", default=env_verbose_level,
    help="the selected test scenario")
args = parser.parse_args()

# alt: format="%(asctime)-23s %(levelname)s - %(name)s: %(message)s")
logging.basicConfig(level={
                           0: logging.WARNING,
                           1: logging.INFO,
                           2: logging.DEBUG
                          }.get(args.verbose, logging.DEBUG),
                    format="%(levelname)s - %(message)s")

ts = SCENARIOS[args.scenario](NDClient(args.zmq), args.library,
                              param=args.param)
try:
    data = ts.run(args.runs)
    if args.file is not None:
        json.dump({"run_items": ts.get_run_items(),
                   "run_label": ts.get_run_label(),
                   "run_data": data,
                   "verify_gs_failed": ts.verify_gs_failed,
                   "beacons": ts.beacons,
                   "failed_beacons": ts.failed_beacons,
                   "requested_predictions": ts.requested_predictions,
                   "failed_predictions": ts.failed_predictions}, args.file)
except KeyboardInterrupt:
    pass
finally:
    if args.file is not None:
        args.file.close()
