#!/usr/bin/env python

import argparse
import json
import logging
import math
import numpy
import matplotlib
from evaluation.prediction_comparator import PredictionComparator
from evaluation.location_comparator import LocationComparator
from evaluation.pattern_comparator import PatternComparator

STATS = {
    "prediction": {
        "comparator": PredictionComparator,
        "reduce": [0, 1],
        "names": ["Contact Start", "Contact End"],
        "ylabel": "diff/s"
    },
    "reliability": {
        "comparator": PredictionComparator,
        "reduce": [2],
        "names": ["Reliability"],
        "ylabel": "diff"
    },
    "capacity": {
        "comparator": PredictionComparator,
        "reduce": [3, 4],
        "names": ["Bitrate Mean", "Bitrate Std. Dev."],
        "ylabel": "diff/bytes"
    },
    "location": {
        "comparator": LocationComparator,
        "reduce": None,
        "names": ["Latitude", "Longitude"],
        "ylabel": "diff/deg"
    },
    "pattern": {
        "comparator": PatternComparator,
        "reduce": None,
        "names": ["Offset", "On-Time", "Duration"],
        "ylabel": "diff/s"
    },
    "ontime": {
        "comparator": PatternComparator,
        "reduce": [1],
        "names": ["On-Time"],
        "ylabel": "diff/s"
    }
}


def run():
    parser = argparse.ArgumentParser(description="Evaluate test results.")
    parser.add_argument("FILE", type=argparse.FileType("r"),
                        help="the file to be read")
    parser.add_argument("-s", "--stat",
                        choices=STATS.keys(), default="prediction",
                        help="the statistic to be generated")
    parser.add_argument("-o", "--over", default="runs",
                        choices=["runs", "gs", "contacts", "all"],
                        help="the entity to generate a stat over")
    parser.add_argument("-p", "--plot", action="store_true",
                        help="generate a plot of the results")
    parser.add_argument("-r", "--relative", action="store_true",
                        help="use relative differences")
    parser.add_argument("-n", "--nan", action="store_true",
                        help="leave NaNs")
    parser.add_argument("--backend", default="Qt5Agg",
                        choices=matplotlib.rcsetup.interactive_bk,
                        help="the matplotlib backend to be used")
    parser.add_argument("-v", "--verbose", action="count", default=0,
                        help="the selected test scenario")
    args = parser.parse_args()
    logging.basicConfig(level={0: logging.INFO}.get(
        args.verbose, logging.DEBUG),
        format="%(asctime)-23s %(levelname)s - %(name)s: %(message)s")
    logger = logging.getLogger(__name__)
    try:
        data = json.load(args.FILE)
    finally:
        args.FILE.close()
    contacts = []
    for run in data["run_data"]:
        for _, gs in run.items():
            for start, end in gs["contacts"]:
                contacts.append((end - start) / 1000.0)
    logger.info("Found {} contacts".format(len(contacts)))
    logger.info("Durations: mean={}s, std={}s, min={}s, max={}s".format(
        round(numpy.mean(contacts), 4), round(numpy.std(contacts), 4),
        numpy.min(contacts), numpy.max(contacts)))
    stat = STATS[args.stat]
    comparator = stat["comparator"](data["run_data"])
    stats = comparator.compare(over=args.over, reduce_to=stat["reduce"],
                               relative_difference=args.relative)
    logger.debug("Stats: \n{}".format(
        json.dumps(stats, indent=4, sort_keys=True)))
    if not args.nan:
        stats = [strip_nan(l) for l in stats]
    xitems, xlabel = xitems_and_label(args.over, data, len(stats), comparator)
    stats = [s for (xi, s) in sorted(zip(xitems, stats))]
    xitems = sorted(xitems)
    stats_n = reshape_stats(stats)
    np_result = [{
            "mean": numpy.mean(l, axis=0).tolist(),
            "std": numpy.std(l, axis=0).tolist()
        } for l in stats]
    for i in range(len(np_result)):
        try:
            np_result[i]["item"] = xitems[i]
        except BaseException as e:
            logger.warning("Cannot assign xitem to result {}: {}".format(i, e))
    logger.info("Result: \n{}".format(
        json.dumps(np_result, indent=4, sort_keys=True)))
    if args.plot:
        do_plot(args.backend, stats_n, xitems, xlabel,
                stat["ylabel"], stat["names"])


def has_no_nan(e):
    for i in e:
        if math.isnan(i):
            return False
    return True


def strip_nan(data):
    return [e for e in data if has_no_nan(e)]


# This splits the innermost results, e.g. for location into lat and lon results
def reshape_stats(stats):
    stats_n = []
    for line_index in range(len(stats)):
        line = stats[line_index]
        for sample in line:
            for column_index in range(len(sample)):
                while column_index >= len(stats_n):
                    stats_n.append([])
                while line_index >= len(stats_n[column_index]):
                    stats_n[column_index].append([])
                stats_n[column_index][line_index].append(sample[column_index])
    return stats_n


def do_plot(backend, stats_n, x_labels, x_name, y_name, subplot_titles):
    matplotlib.use(backend)
    import matplotlib.pyplot as plt
    cols = len(stats_n)
    fig, axis = plt.subplots(ncols=cols)
    if cols == 1:
        axis = [axis]
    for i in range(len(stats_n)):
        a = axis[i]
        a.boxplot(stats_n[i], whis=[5, 95], showfliers=False)
        if len(str(x_labels[0])) > 3:
            a.set_xticklabels(x_labels, rotation=45,
                              horizontalalignment="right")
        elif len(x_labels) > 15:
            import matplotlib.ticker as ticker
            a.xaxis.set_major_locator(ticker.MultipleLocator(5))
            a.set_xticklabels(x_labels[0::5])
        else:
            a.set_xticklabels(x_labels)
        a.set_xlabel(x_name)
        a.set_ylabel(y_name)
        a.set_title(subplot_titles[i])
    plt.tight_layout(w_pad=2)
    plt.show()


def xitems_and_label(over, data, stat_count, comparator):
    xitems = range(stat_count)
    xlabel = ""
    if over == "gs":
        xitems = comparator.eids
    elif over == "runs":
        xitems = data["run_items"]
        xlabel = data["run_label"]
    elif over == "contacts":
        xlabel = "contact"
    return xitems, xlabel

if __name__ == "__main__":
    run()
