import abc
import numpy
import logging

logger = logging.getLogger(__name__)


class StatisticalComparator(object, metaclass=abc.ABCMeta):

    def __init__(self, data):
        self.configured = []
        self.measured = []
        assert len(data) > 0
        self.eids = sorted(data[0].keys())
        for run in data:
            conf_run = []
            meas_run = []
            assert len(run.keys()) == len(self.eids)
            for eid in self.eids:
                gs = run[eid]
                assert len(gs["contacts"]) <= len(gs["observations"])
                conf_gs, meas_gs = self._gsrun_data_to_tuple(gs)
                assert len(conf_gs) == len(meas_gs)
                conf_run.append(conf_gs)
                meas_run.append(meas_gs)
            assert len(conf_run) > 0
            self.configured.append(conf_run)
            self.measured.append(meas_run)
        assert len(self.configured) > 0

    @abc.abstractmethod
    def _gsrun_data_to_tuple(self, data):
        pass

    def compare(self, over="runs", relative_difference=True, reduce_to=None):
        diff = (StatisticalComparator._relative_difference
                if relative_difference else StatisticalComparator._difference)
        if over == "runs":
            stats = StatisticalComparator._stats_per_run
        elif over == "gs":
            stats = StatisticalComparator._stats_per_gs
        elif over == "contacts":
            stats = StatisticalComparator._stats_per_contact
        elif over == "all":
            stats = StatisticalComparator._stats_all
        else:
            raise Exception(
                "over has to be one of 'runs', 'gs', 'contacts', 'all")
        return stats(self.configured, self.measured,
                     lambda c, m: StatisticalComparator._reduce_to(diff(c, m),
                                                                   reduce_to))

    @staticmethod
    def _stats_per_run(c, m, diff):
        result = []
        for i_run in range(len(c)):
            l = []
            for i_gs in range(len(c[i_run])):
                for i_ct in range(len(c[i_run][i_gs])):
                    l.append(diff(c[i_run][i_gs][i_ct], m[i_run][i_gs][i_ct]))
            result.append(l)
        return result

    @staticmethod
    def _stats_per_gs(c, m, diff):
        result = []
        for i_gs in range(len(c[0])):
            l = []
            for i_run in range(len(c)):
                assert len(c[i_run]) == len(c[0])
                for i_ct in range(len(c[i_run][i_gs])):
                    l.append(diff(c[i_run][i_gs][i_ct], m[i_run][i_gs][i_ct]))
            result.append(l)
        return result

    @staticmethod
    def _stats_per_contact(c, m, diff):
        contacts = max([len(l) for l in c[0]])
        result = []
        for i_ct in range(contacts):
            l = []
            for i_run in range(len(c)):
                for i_gs in range(len(c[i_run])):
                    if i_ct < len(c[i_run][i_gs]):
                        l.append(diff(c[i_run][i_gs][i_ct],
                                      m[i_run][i_gs][i_ct]))
            result.append(l)
        return result

    @staticmethod
    def _stats_all(c, m, diff):
        l = []
        for i_run in range(len(c)):
            for i_gs in range(len(c[i_run])):
                for i_ct in range(len(c[i_run][i_gs])):
                    l.append(diff(c[i_run][i_gs][i_ct],
                                  m[i_run][i_gs][i_ct]))
        return [l]

    @staticmethod
    def _relative_difference(configured, measured):
        if measured is None:
            return [float('nan')]
        if measured[0] is None:
            return [float('nan')] * len(measured)
        try:
            return numpy.divide(
                numpy.subtract(measured, configured), configured).tolist()
        except BaseException as e:
            logger.debug("Can't calculate difference of {} and {}: {}"
                         .format(configured, measured, e))
            return [float('nan')]

    @staticmethod
    def _difference(configured, measured):
        if measured is None:
            return [float('nan')]
        if measured[0] is None:
            return [float('nan')] * len(measured)
        try:
            return numpy.subtract(measured, configured).tolist()
        except BaseException as e:
            logger.debug("Can't calculate difference of {} and {}: {}"
                         .format(configured, measured, e))
            return [float('nan')]

    @staticmethod
    def _reduce_to(tup, reduce_to=None):
        logger.debug("Reducing  {}".format(tup))
        if type(tup) is not tuple and type(tup) is not list:
            return [tup]
        if reduce_to is None:
            res = list(tup)
        else:
            res = [tup[i] for i in reduce_to]
        return res
