import numpy
from math import sqrt, sin, cos, atan2
from skyfield.api import Angle
from logging import getLogger

logger = getLogger(__name__)


def difference(configured, measured):
    # return numpy.round(numpy.divide(
    #     numpy.subtract(measured, configured), configured), 6).tolist()
    return numpy.round(numpy.subtract(measured, configured), 6).tolist()


class AvailabilityPattern(object):

    def __init__(self, offset, on_time, period):
        assert period != 0
        assert on_time <= period
        assert offset < period
        self.duration = period
        self.offset = offset % period
        self.on_time = on_time

    def is_available(self, time):
        return self.get_available_time(time) <= self.on_time

    def get_available_time(self, time):
        return (time - self.offset) % self.duration

    def compare(self, pattern):
        assert isinstance(pattern, AvailabilityPattern)
        return (difference(self.offset, pattern.offset),
                difference(self.on_time, pattern.on_time),
                difference(self.duration, pattern.duration))


class GroundStation(object):

    def __init__(self, eid, coords, reliability=1.0, beacon_period=10,
                 min_altitude=0.0, tx_bitrate=9600, rx_bitrate=9600,
                 availability_pattern=None,
                 achieved_rx_mean=None, achieved_rx_std=0):
        assert len(eid) > 0
        assert len(coords) == 2
        assert reliability >= 0.0 and reliability <= 1.0
        assert beacon_period > 0 and beacon_period <= 60
        assert min_altitude >= 0.0
        if availability_pattern is not None:
            assert isinstance(availability_pattern, AvailabilityPattern)
        self.eid = eid
        self.latitude = Angle(degrees=coords[0])
        self.longitude = Angle(degrees=coords[1])
        self.reliability = reliability
        self.beacon_period = beacon_period
        self.min_altitude = min_altitude
        self.tx_bitrate = tx_bitrate
        self.rx_bitrate = rx_bitrate
        self.achieved_rx = (achieved_rx_mean
                            if achieved_rx_mean is not None else rx_bitrate)
        self.achieved_rx_std = achieved_rx_std
        self.pattern = availability_pattern
        self.verified_once = False
        self.last_prediction_in_the_future = False
        self.data = {
            "eid": eid,
            "coords": coords,
            "reliability": reliability,
            "beacon_period": beacon_period,
            "min_altitude": min_altitude,
            "tx_bitrate": tx_bitrate,
            "rx_bitrate": rx_bitrate,
            "achieved_rx": [self.achieved_rx, self.achieved_rx_std],
            "pattern": ((availability_pattern.offset,
                         availability_pattern.on_time,
                         availability_pattern.duration)
                        if availability_pattern is not None else None),
            "contacts": [],
            "observations": []
        }

    def verify_and_store(self, response, next_start=None,
                         last_contact=(None, None), before_contact=None):
        entry = {}
        try:
            entry = self.verify_response(response, next_start, last_contact,
                                         before_contact)
        finally:
            assert len(self.data["observations"]) == int(before_contact)
            self.data["observations"].append(entry)

    def verify_response(self, response, next_start=None,
                        last_contact=(None, None), before_contact=None):
        assert (response["result"] & 128) == 0
        gs = response["gs"]
        assert gs["eid"] == self.eid
        assert gs["beacon_period"] / 100 == self.beacon_period
        # assert gs["tx_bitrate"] / 8 == self.tx_bitrate
        # assert gs["rx_bitrate"] / 8 == self.rx_bitrate
        reliability_diff = difference(self.reliability,
                                      gs["reliability"])
        lat, lon, alt = GroundStation.ecef2lla(*gs["location"])
        alt /= 1000.0
        topo_difference = difference(
            [self.latitude.radians, self.longitude.radians, 0],
            [lat, lon, abs(alt)])
        if self.pattern is not None:
            try:
                pattern_meas = AvailabilityPattern(
                    *[x // 1000 for x in gs["ap"]])
                pattern_diff = self.pattern.compare(pattern_meas)
            except:
                pattern_diff = None
        else:
            pattern_diff = None
        if last_contact[0] is not None and last_contact[1] is not None:
            contact_diff = difference(
                list(last_contact),
                numpy.divide([gs["last_contact_start_time"],
                              gs["last_contact_end_time"]],
                             1000.0))
        else:
            contact_diff = None
        if gs["next_prediction"][0] != 0 and next_start is not None:
            # TODO: Accumulate predictions and check them at the end!
            logger.debug("GS {} - Prediction: {}".format(
                self.eid, gs["next_prediction"]))
            prediction_diff = int(next_start - gs["next_prediction"][0] / 1000)
            if prediction_diff < -1800:
                self.last_prediction_in_the_future = True
            else:
                self.last_prediction_in_the_future = False
            reliability_diff = difference(self.reliability,
                                          gs["next_prediction"][2])
        else:
            prediction_diff = None
        diffs = (reliability_diff, topo_difference, pattern_diff,
                 contact_diff, prediction_diff)
        logger.info("GS {} - DIFF: R={}, T={}, A={}(s), C={}(s), P={}(s)"
                    .format(self.eid, *diffs))
        self.verified_once = True
        if before_contact is not None:
            return {
                "coords": (lat, lon),
                "prediction": (gs["next_prediction"]
                               if prediction_diff is not None else None),
                "pattern": ((pattern_meas.offset, pattern_meas.on_time,
                             pattern_meas.duration)
                            if pattern_diff is not None else None)
            }
        return None

    def add_contact(self, contact, number):
        assert len(self.data["contacts"]) == int(number)
        self.data["contacts"].append((contact[0] * 1000, contact[1] * 1000))

    def get_collected_data(self):
        return self.data

    # From geodesy JS library by Chris Veness, MIT licensed
    # https://github.com/chrisveness/geodesy/blob/master/latlon-ellipsoidal.js
    @staticmethod
    def ecef2lla(x, y, z):
        try:
            # WGS84
            a = 6378.137
            b = 6356.75231425

            e2 = (a * a - b * b) / (a * a)    # 1st eccentricity squared
            eps2 = (a * a - b * b) / (b * b)  # 2nd eccentricity squared
            p = sqrt(x * x + y * y)           # distance from minor axis
            R = sqrt(p * p + z * z)           # polar radius

            # parametric latitude
            # (Bowring eqn 17, replacing tan_beta = z·a / p·b)
            tan_beta = (b * z) / (a * p) * (1 + eps2 * b / R)
            sin_beta = tan_beta / sqrt(1 + tan_beta * tan_beta)
            cos_beta = sin_beta / tan_beta

            # geodetic latitude (Bowring eqn 18)
            phi = atan2(z + eps2 * b * sin_beta * sin_beta * sin_beta,
                        p - e2 * a * cos_beta * cos_beta * cos_beta)

            # longitude
            lamb = atan2(y, x)

            # height above ellipsoid (Bowring eqn 7)
            sin_phi = sin(phi)
            cos_phi = cos(phi)
            # length of the normal terminated by the minor axis
            length = a / sqrt(1 - e2 * sin_phi * sin_phi)
            h = p * cos_phi + z * sin_phi - (a * a / length)

            return (phi, lamb, h)
        except:
            return (float('nan'), float('nan'), float('nan'))
