import abc
import sys
import traceback
import random
import ephem
import math
import numpy
from datetime import datetime, timezone
from skyfield.api import load, JulianDate
from logging import getLogger, DEBUG
from .status import RRNDStatus

logger = getLogger(__name__)


class TestScenario(object, metaclass=abc.ABCMeta):

    RAD2DEG = 180.0 / math.pi

    def __init__(self, client, library, ephemeris="de421.bsp", param=None):
        self.client = client
        self.library = library
        self.ephemeris = ephemeris
        self.param = param
        self.log_beacons = (logger.getEffectiveLevel() <= DEBUG)
        self.verify_gs_failed = 0
        self.beacons = 0
        self.failed_beacons = 0
        self.requested_predictions = 0
        self.failed_predictions = 0

    @abc.abstractmethod
    def _get_scenario(self, run):
        pass

    @abc.abstractmethod
    def _get_default_runs(self):
        pass

    def run(self, runs=0):
        if not self.client.is_connected():
            logger.error("Could not connect to ZMQ")
            raise Exception()
        logger.info("Using the {} library".format(self.library))
        logger.info("Running scenario '{}'".format(type(self).__name__))
        if self.library == "skyfield":
            logger.info("Using ephemeris data '{}'".format(self.ephemeris))
            planets = load(self.ephemeris)
            self.earth = planets["earth"]
        else:
            self.earth = None
        if runs <= 0:
            runs = self._get_default_runs()
        if not self.client.send_resetperf():
            logger.error("Could not reset perf data")
            raise Exception()
        self.verify_gs_failed = 0
        self.beacons = 0
        self.failed_beacons = 0
        self.requested_predictions = 0
        self.failed_predictions = 0
        data = []
        for run in range(runs):
            data.append(self._do_run(run))
        logger.info("Scenario finished: {} failed gs, {} beacons, {} failed"
                    .format(self.verify_gs_failed, self.beacons,
                            self.failed_beacons))
        logger.info("Predictions requested: {}, failed: {}"
                    .format(self.requested_predictions,
                            self.failed_predictions))
        return data

    def _do_run(self, run):
        data = self._prepare_run(self.earth, self._get_scenario(run), run)
        self._single_run(*data)
        if not self.client.send_storeperf():
            logger.warning("Could not store perf data")
        return {eid: v[0].get_collected_data() for eid, v in data[1].items()}

    def _create_obs(self, lat, lon, earth):
        if self.library == "skyfield":
            obs = earth.topos(lat, lon)
        else:
            obs = ephem.Observer()
            obs.lat = str(lat.degrees)
            obs.lon = str(lon.degrees)
        return obs

    def _prepare_run(self, earth, sd, run):
        if self.library == "skyfield":
            sat = earth.satellite(sd["tle"])
            t0 = JulianDate(utc=sd["t0"]).tai
            alt = TestScenario._skyfield_altitude
            inc_jd = True
        else:
            sat = ephem.readtle(*sd["tle"].strip().split("\n"))
            t0 = datetime(*sd["t0"], tzinfo=timezone.utc)
            alt = TestScenario._ephem_altitude
            inc_jd = False
        gs = {s.eid: (s,      # instance
                      self._create_obs(s.latitude, s.longitude, earth),
                      None,   # last contact start
                      None,   # last contact end
                      False,  # in contact?
                      False,  # inactive contact?
                      0)      # no. of contact
              for s in sd["ground_stations"]}
        satname = sd["tle"].strip().split("\n")[0].strip()
        logger.info("Run {} with sat '{}' and {} ground station(s): {}".format(
            run, satname, len(gs), list(gs.keys())))
        unixtime = datetime(1970, 1, 1, tzinfo=timezone.utc)
        start_time = int(round((datetime(*sd["t0"], tzinfo=timezone.utc) -
                                unixtime).total_seconds(), 0))
        end_time = start_time + int(86400 * sd["duration"])
        if not self.client.send_reset():
            logger.error("Could not reset RRND")
            raise Exception()
        if not self.client.send_tle(sd["tle"], start_time):
            logger.error("Could not initialize satpos by TLE")
            raise Exception()
        return [sat, gs, start_time, end_time, int(sd["period"]),
                t0, t0, alt, inc_jd]

    def _single_run(self, sat, gs, start_time, end_time, period, t0,
                    jd, alt, inc_jd):
        beacon_seq = {eid: 0 for eid in gs.keys()}
        jd_step = float(period) / 24 / 60 / 60
        for ts in range(start_time, end_time, period):
            for eid, (s, topos, c_start, c_end, in_c, inact, no) in gs.items():
                c = (alt(topos, sat, ts, jd) > s.min_altitude)
                c &= (s.pattern is None or s.pattern.is_available(ts))
                if c != in_c:
                    logger.info("Contact {} for gs {} @ {}".format(
                        "started" if c else "ended", eid, ts))
                    if c:
                        self.verify_gs(s, ts - 300, ts, (c_start, c_end), no)
                        c_start = ts
                        inact = (random.random() > s.reliability)
                        if inact:
                            logger.info("Contact with {} will be missed"
                                        .format(eid))
                    else:
                        c_end = ts - period
                        s.add_contact((c_start, c_end), no)
                        no += 1
                        if self.log_beacons:
                            print("")
                    in_c = c
                    gs[eid] = (s, topos, c_start, c_end, in_c, inact, no)
                if c and not inact:
                    self.send_beacon(
                        beacon_seq[eid], eid, ts, period, s.pattern,
                        s.tx_bitrate, s.achieved_rx, s.achieved_rx_std)
                    beacon_seq[eid] += 1
            if inc_jd:
                jd += jd_step
        logger.info("Finished")
        for eid, (s, topos, c_start, c_end, in_c, inact, no) in gs.items():
            self.verify_gs(s)

    def send_beacon(self, seq, eid, unix_timestamp, period, avail_p=None,
                    br_tx=9600, br_rx=9600, br_rx_std=0):
        avail = (True if avail_p is None
                 else avail_p.get_available_time(unix_timestamp))
        rx = (br_rx if br_rx_std == 0
              else numpy.random.normal(br_rx, br_rx_std))
        if rx < 0.2:
            rx = 0.2
        success, code = self.client.send_beacon(seq, eid, unix_timestamp,
                                                period, availability=avail,
                                                bitrate_up=br_tx,
                                                bitrate_down=rx)
        # if not success:
        #     logger.warning("Failed sending beacon: {}".format(hex(code)))
        self.beacons += 1
        if not success:
            self.failed_beacons += 1
        if not self.log_beacons:
            return
        if not success:
            print("!", end="")
        elif code == RRNDStatus.UPDATED:
            print("_", end="")
        elif (code & RRNDStatus.DELETE) != 0:
            print("~", end="")
        elif (code & RRNDStatus.UPDATE_LOCATOR_DATA) != 0:
            print("#", end="")
        elif (code & RRNDStatus.UPDATE_AVAILABILITY) != 0:
            print(":", end="")
        elif (code & RRNDStatus.UPDATE_CONTACTS) != 0:
            print("+", end="")
        else:
            print("?", end="")

    def verify_gs(self, gs, time=None, ptime=None, last_contact=(None, None),
                  counter=None):
        try:
            if time is not None:
                if not gs.last_prediction_in_the_future:
                    # TODO: Check last p-time against current one!
                    self.requested_predictions += 1
                    ret = self.client.request_prediction(gs.eid, time)
                    if (ret is None or "result" not in ret or
                            (ret["result"] & RRNDStatus.FAILED) != 0):
                        self.failed_predictions += 1
            else:
                ptime = None
            gsdata = self.client.query_gs(gs.eid)
            if counter is not None:
                gs.verify_and_store(gsdata, ptime, last_contact, counter)
            else:
                gs.verify_response(gsdata, ptime, last_contact, counter)
            return gsdata
        except (AssertionError, KeyError):
            if gs.verified_once:
                logger.warning("Could not verify GS {}: {} failed".format(
                    gs.eid, traceback.extract_tb(sys.exc_info()[2])[-1][3]))
                self.verify_gs_failed += 1

    @staticmethod
    def _skyfield_altitude(topos, sat, ts, jd):
        return topos.at(tai=jd).observe(sat).altaz()[0].degrees

    @staticmethod
    def _ephem_altitude(topos, sat, ts, jd):
        topos.date = datetime.utcfromtimestamp(ts)
        sat.compute(topos)
        return sat.alt * TestScenario.RAD2DEG

    def get_run_items(self):
        try:
            return self.RUN_ITEMS
        except AttributeError:
            return list(range(self._get_default_runs()))

    def get_run_label(self):
        try:
            return self.RUN_LABEL
        except AttributeError:
            return "run"
