/* System includes */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/* uPCN includes */
#include <upcn/upcn.h>
#include <upcn/bundle.h>

/* Test includes */
#include <upcnTest.h>
#include <testlib.h>

#define BUNDLE_PAYLOAD_SIZE 500
#define BUNDLE_DEST "dtn:bundle.receiver"

#define TEST_BEGIN 80000
#define CONTACT_BEGIN 80003
#define CONTACT_DURATION 3
#define CONTACT_SPACING 6
#define CONTACT_PERIOD 20
#define T_BEACON_PERIOD 1
#define BANDWIDTH 800 /* 9600 Baud => 2/3 */

#define TEST_ITERATIONS 1
#define ITER_CONTACTS 9

#define EP_1 "upcn:ahn.1.ep"
#define EP_2 "upcn:ahn.2.ep"
#define EP_3 "upcn:web.ep"

#define CONTACT_COUNT 9
#define CONTACT_MAX_BDL 7

#define COOKIE_LEN (UPCN_HASH_LENGTH + 4)

static char *bdldests[3] = {
	EP_1,
	EP_2,
	EP_3
};

static int out_bdl_c;
static struct bundle *out_bdl[CONTACT_COUNT * CONTACT_MAX_BDL];

static int contact_bdl_c[CONTACT_COUNT];
static struct bundle *contact_bdl[CONTACT_COUNT][CONTACT_MAX_BDL];

static struct beacon *own_b1[3], *own_b2[3];

void drt_init(void)
{
	int i;

	for (i = 0; i < 3; i++) {
		own_b1[i] = upcntest_create_beacon(bdldests[i],
			T_BEACON_PERIOD, BANDWIDTH,
			NULL, 0, (const char **)&bdldests[i], 1);
		own_b2[i] = upcntest_create_beacon(bdldests[i],
			T_BEACON_PERIOD, BANDWIDTH,
			malloc_unmanaged(COOKIE_LEN), COOKIE_LEN,
			(const char **)&bdldests[i], 1);
	}
}

void drt_cleanup_lists(void)
{
	int i, c;

	for (i = 0; i < out_bdl_c; i++) {
		if (out_bdl[i] != NULL) {
			bundle_free(out_bdl[i]);
			out_bdl[i] = NULL;
		}
	}
	out_bdl_c = 0;
	for (i = 0; i < CONTACT_COUNT; i++) {
		for (c = 0; c < contact_bdl_c[i]; c++) {
			if (contact_bdl[i][c] != NULL) {
				bundle_free(contact_bdl[i][c]);
				contact_bdl[i][c] = NULL;
			}
		}
		contact_bdl_c[i] = 0;
	}
}

void drt_cleanup(void)
{
	int i;

	drt_cleanup_lists();
	for (i = 0; i < 3; i++) {
		beacon_free(own_b1[i]);
		beacon_free(own_b2[i]);
	}
}

static uint64_t contacttime(int contact)
{
	int orbit = contact / 3;
	uint64_t orbit_begin = CONTACT_BEGIN + CONTACT_PERIOD * orbit;

	return orbit_begin + (contact % 3) * CONTACT_SPACING;
}

/* Discovery Test */

static void beacsend(struct beacon *b)
{
	size_t s = 0;
	uint8_t *bin = upcntest_beacon_to_bin(b, &s, 0);

	printf("Sending beacon...\n");
	upcntest_begin_send(INPUT_TYPE_BEACON_DATA);
	upcntest_send_binary(bin, s);
	upcntest_finish_send();
}

static uint8_t *beacrcv(void)
{
	struct beacon *bc = upcntest_receive_next_beacon(500);
	struct tlv_definition *td;
	uint8_t *ret;

	if (bc == NULL)
		return NULL;
	printf("Received beacon: ");
	td = beacon_get_service(bc, TLV_TYPE_PRIVATE_COOKIES);
	if (td == NULL) {
		printf("no cookie!\n");
		beacon_free(bc);
		return NULL;
	}
	td = tlv_get_child(td, TLV_TYPE_BYTES);
	if (td == NULL) {
		printf("no cookie bytes!\n");
		beacon_free(bc);
		return NULL;
	}
	if (td->length != COOKIE_LEN) {
		if (td->length != 0 && td->length != 1)
			printf("wrong HASH length %d!\n", td->length);
		else
			printf("empty cookie!\n");
		beacon_free(bc);
		return NULL;
	}
	printf("cookie OK!\n");
	ret = malloc(COOKIE_LEN);
	memcpy(ret, td->value.b, COOKIE_LEN);
	beacon_free(bc);
	return ret;
}

static void send_contact_beacons(int contact)
{
	uint64_t contact_begin = contacttime(contact);
	uint64_t next_beacon = contact_begin;
	int beacon_count = CONTACT_DURATION / T_BEACON_PERIOD;
	int gs = contact % 3;
	int orbit = contact / 3;
	int c;
	struct beacon *b;

	if (orbit == 0)
		b = own_b1[gs];
	else
		b = own_b2[gs];
	for (c = 0; c < beacon_count; c++) {
		wait_until(next_beacon);
		beacsend(b);
		next_beacon += T_BEACON_PERIOD;
	}
}

static void add_cookie_to_own(uint8_t *cookie, int gs)
{
	struct tlv_definition *td;

	td = beacon_get_service(own_b2[gs], TLV_TYPE_PRIVATE_COOKIES);
	ASSERT(td != NULL);
	td = tlv_get_child(td, TLV_TYPE_BYTES);
	ASSERT(td != NULL);
	ASSERT(td->length == COOKIE_LEN);
	memcpy(td->value.b, cookie, COOKIE_LEN);
	free(cookie);
}

/* Routing Test */

static void send_with_prio(int dest, int high)
{
	size_t s = 0;
	uint8_t *bin;
	struct bundle *b;

	b = upcntest_create_bundle(BUNDLE_PAYLOAD_SIZE, bdldests[dest], 1, 0);
	b->proc_flags &= ~(BUNDLE_FLAG_NORMAL_PRIORITY
		| BUNDLE_FLAG_EXPEDITED_PRIORITY);
	b->proc_flags |= (high ? BUNDLE_FLAG_EXPEDITED_PRIORITY
		: BUNDLE_FLAG_NORMAL_PRIORITY);
	out_bdl[out_bdl_c++] = b;
	bin = upcntest_bundle_to_bin(b, &s, 0);
	upcntest_begin_send(INPUT_TYPE_BUNDLE_DATA);
	upcntest_send_binary(bin, s);
	upcntest_finish_send();
}

/* TO 1 LO | TO 1 HI | ... | TO 3 HI | FROM LO | FROM HI */
static const int contact_mapping[CONTACT_COUNT][8] = {
	{0, 0, 0, 0, 2, 2, 0, 0},
	{0, 0, 0, 0, 2, 1, 0, 0},
	{1, 1, 0, 1, 0, 0, 2, 3},
	{0, 0, 0, 0, 0, 1, 1, 1},
	{0, 0, 0, 0, 0, 0, 0, 1},
	{1, 2, 1, 0, 0, 0, 5, 1},
	{0, 0, 0, 0, 0, 0, 1, 2},
	{0, 0, 0, 0, 0, 0, 1, 0},
	{0, 0, 0, 0, 0, 0, 0, 0}
};

static int bprio(struct bundle *b)
{
	return (b->proc_flags & BUNDLE_FLAG_EXPEDITED_PRIORITY) != 0;
}

static int handle_contact(int c)
{
	int i, j, n, p, b = 0, firstc = ITER_CONTACTS, success = 1;

	wait_until(contacttime(firstc + c));
	printf("Handling contact %d...\n", c);
	for (i = 5; i >= 0; i--)
		for (j = 0; j < contact_mapping[c][i]; j++)
			send_with_prio(i / 2, i % 2);
	for (i = 6; i < 8; i++)
		b += contact_mapping[c][i];
	contact_bdl_c[c] = b;
	if (b > CONTACT_MAX_BDL)
		b = CONTACT_MAX_BDL;
	/* Beacons */
	send_contact_beacons(c + ITER_CONTACTS);
	/* Bundles */
	p = 2;
	for (i = 0; i < b; i++) {
		printf("Waiting for bundle... ");
		contact_bdl[c][i] = upcntest_receive_next_bundle(500);
		if (contact_bdl[c][i] == NULL) {
			success = 0;
			printf("FAIL\n");
			continue;
		}
		if (bprio(contact_bdl[c][i]) > p) {
			success = 0;
			printf("<PRIOFAIL> ");
		}
		p = bprio(contact_bdl[c][i]);
		for (n = 0; n < out_bdl_c; n++) {
			if (bundlecmp(out_bdl[n], contact_bdl[c][i])) {
				printf("= sent bundle #%d ", n);
				break;
			}
		}
		printf("OK\n");
	}
	return success ? EXIT_SUCCESS : EXIT_FAILURE;
}

void drt_run(void)
{
	static uint8_t *cookie;
	int i, c, success = 1;

	upcntest_send_query();
	upcntest_send_settime(TEST_BEGIN);
	resettime(TEST_BEGIN);
	upcntest_send_resetstats();
	printf("First run: Discovery...\n");
	for (i = 0; i < ITER_CONTACTS; i++) {
		send_contact_beacons(i);
		if (i / 3 == 0) {
			/* First orbit: cookie check */
			c = 0;
			cookie = NULL;
			while (c++ != 8 && cookie == NULL)
				cookie = beacrcv();
			ASSERT(cookie != NULL);
			add_cookie_to_own(cookie, i % 3);
		}
	}
	printf("Second run: Routing Test...\n");
	/* Test */
	for (i = 0; i < ITER_CONTACTS; i++)
		if (handle_contact(i) != EXIT_SUCCESS)
			success = 0;
	/* Delete bundles from list, ... */
	drt_cleanup_lists();
	/* Get logfile */
	upcntest_send_storetrace();
	upcntest_send_query();
	ASSERT(success);
}
