/* System includes */
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

/* uPCN includes */
#include <upcn/upcn.h>
#include <upcn/beacon.h>

/* Test includes */
#include <upcnTest.h>
#include <testlib.h>

#define COOKIE_LEN (UPCN_HASH_LENGTH + 4)

#define T_BEACON_PERIOD 1
#define T_CONTACT_BEGIN 1
#define T_CONTACT_DURATION 5
#define T_ORBIT_PERIOD 30
#define T_BR 1200

#define GS "upcn:testgs"
#define EP "upcn:testep"

static struct beacon *own_b1, *own_b2;

void dt_init(void)
{
	static char *ep1 = EP;

	own_b1 = upcntest_create_beacon(GS, T_BEACON_PERIOD, T_BR,
		NULL, 0, (const char **)&ep1, 1);
	own_b2 = upcntest_create_beacon(GS, T_BEACON_PERIOD, T_BR,
		malloc_unmanaged(COOKIE_LEN), COOKIE_LEN,
		(const char **)&ep1, 1);
}

void dt_cleanup(void)
{
	beacon_free(own_b1);
	beacon_free(own_b2);
}

static void beacsend(struct beacon *b)
{
	size_t s = 0;
	uint8_t *bin = upcntest_beacon_to_bin(b, &s, 0);

	printf("Sending beacon...\n");
	upcntest_begin_send(INPUT_TYPE_BEACON_DATA);
	upcntest_send_binary(bin, s);
	upcntest_finish_send();
}

static uint8_t *beacrcv(void)
{
	struct beacon *bc = upcntest_receive_next_beacon(500);
	struct tlv_definition *td;
	uint8_t *ret;

	if (bc == NULL)
		return NULL;
	printf("Received beacon: ");
	td = beacon_get_service(bc, TLV_TYPE_PRIVATE_COOKIES);
	if (td == NULL) {
		printf("no cookie!\n");
		beacon_free(bc);
		return NULL;
	}
	td = tlv_get_child(td, TLV_TYPE_BYTES);
	if (td == NULL) {
		printf("no cookie bytes!\n");
		beacon_free(bc);
		return NULL;
	}
	if (td->length != COOKIE_LEN) {
		if (td->length != 0 && td->length != 1)
			printf("false HASH length %d!\n", td->length);
		else
			printf("empty cookie!\n");
		beacon_free(bc);
		return NULL;
	}
	printf("cookie OK!\n");
	ret = malloc(COOKIE_LEN);
	memcpy(ret, td->value.b, COOKIE_LEN);
	beacon_free(bc);
	return ret;
}

static void send_contact_beacons(struct beacon *b, int contact)
{
	uint64_t contact_begin = T_CONTACT_BEGIN + T_ORBIT_PERIOD * contact;
	uint64_t next_beacon = contact_begin;
	int beacon_count = T_CONTACT_DURATION / T_BEACON_PERIOD;
	int c;

	for (c = 0; c < beacon_count; c++) {
		wait_until(next_beacon);
		beacsend(b);
		next_beacon += T_BEACON_PERIOD;
	}
}

static void add_cookie_to_own(uint8_t *cookie)
{
	struct tlv_definition *td;

	td = beacon_get_service(own_b2, TLV_TYPE_PRIVATE_COOKIES);
	ASSERT(td != NULL);
	td = tlv_get_child(td, TLV_TYPE_BYTES);
	ASSERT(td != NULL);
	ASSERT(td->length == COOKIE_LEN);
	memcpy(td->value.b, cookie, COOKIE_LEN);
	free(cookie);
}

void dt_run(void)
{
	int i;
	uint8_t *cookie;

	upcntest_send_query();
	upcntest_send_resetstats();
	resettime(0);
	send_contact_beacons(own_b1, 0);
	i = 0;
	cookie = NULL;
	while (i++ != 8 && cookie == NULL)
		cookie = beacrcv();
	ASSERT(cookie != NULL);
	add_cookie_to_own(cookie);
	send_contact_beacons(own_b2, 1);
	send_contact_beacons(own_b2, 2);
	upcntest_send_storetrace();
	upcntest_send_query();
	upcntest_send_router_query();
}
